#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def semivariogram(x, y, v, nL, di, td, stype='omnidirectional', negscat=0.,
                  model='exponential', graph=False, lunit='m', p0=(1.,1.,1.),
                  runtimediag=False):
    """
        Calculates single or multiple experimental semivariogram(s) for
        spatial distributed data. X and Y coordinates and corresponding
        values are required in separate numpy arrays. Different theoretical
        semivariograms can be fitted or the best fitting is chosen.

        Results can be plotted in various graphs and the fitted model parameters
        are given to the output.

        Small code parts are based on Alghalandis.com


        Definition
        ----------
        def semivariogram(x,y,v,nL,di,td,stype='omnidirectional',negscat=0.,
                          model='exponential',graph=False,lunit='m',
                          p0=(0.5,0.5,100),runtimediag=False):


        Input
        -----
        x         numpy array with longitude ('easting')
        y         numpy array with latitude ('northing')
        v         numpy array with values
        nL        int with number of lags to calculate
        di        list with angles to consider (in degree). di's > 180
                  or < -180 are not allowed.
        td        int with one sided angle tolerance, so td=30 means
                  an angular span of 60 (in degree). Choose td big
                  enough, so that no "empty" angular span with no
                  samples appear. Otherwise, an exeption is raised.
                  td > 180 is not allowed.


        Optional Input
        --------------
        stype      type of semivariogram

            'omnidirectional' semivariogram is calculated for the hole
                              circle from 0 to 180 and 0 to -180. So
                              every direction is considered. (default)
            'directional'     semivariogram is calculated only for the
                              given angles in di. Orientation is not
                              considered, so that e.g. 45 deg equals
                              -135 deg. Due to that, negative angles
                              in di make no sense (and are not allowed)
                              here.
            'directional+orientational'
                              semivariogram is calculated only for the
                              given angles in di. Orientation is
                              considered. Every angle between 0 to 180
                              and 0 to -180 are allowed.

        negscat   It is common in geostatistics to neglect large lags which
                  tend to scatter and contain outliers. By setting negscat to
                  a value between 0 and 1 (both excluded) you can cut off the
                  tail of the experimental semivariogram and the fitting will
                  be done only to the remaining part. If it's set to False, the
                  hole semivariogram is calculated. (default=False)

        model     type of theoretical semivariogram model to fit to the
                  experimental semivariogram.

            'exponential'     exponential semivariogram model (default)
            'spherical'       sperical semivariogram model
            'gaussian'        gaussian semivariogram model
            'power'           power semivariogram model
            'noidea'          fitts all three models and returns the one
                              with the smallest coefficient of variation
                              (That does not mean, the returning model
                              is the best model for your data, it only
                              has the smallest parameter variation, but
                              it may help you with the decision for the
                              right model)
            'nomodel'         model fitting is disabled

        graph     If True, plotting is enabled. If False, plotting is
                  disabled. (default=True)

        lunit     Unit of the longitude and latitude (default=meter,
                  only used for plot labeling)

        p0        Initial guess for parameter estimation of the
                  theoretical semivariogram.
                  1st entry: equals approximately the nugget
                  2nd entry: equals approximately nugget-sill
                  3rd entry: equals approximately the range
                  (default=(0.5,0.5,100))
                  Sometimes you have to play around with it if no
                  fitting result can be reached.

        runtimediag
                  If True, during run time of the function, diagnostics
                  are printed to the console. If False, runtime diagnostics
                  print is disabled. (default=True)


        Output
        ------
        if model is set to 'nomodel':
        h         list(s) of lags
        g         list(s) of variances
        c         list(s) of samples per lag

        otherwise:
        nugget    height of nugget(s) [(unit of v)**2], always >=0
        sill      height if sill(s) [(unit of v)**2], always >=0
        irange     distance of range [unit of x and y], always >=0
        vark      coefficient(s) of variation averaged over all three
                  model parameter (goodness-of-fit) [-]
        pcov      matrices of parameter variance and covariances
                  of the model parameters
        h         array(s) of lags
        g         array(s) of variances
        c         array(s) of samples per lag
        mod       model function, either the one you selected or the model with
                  the best fit if you selected model='noidea'
        popt      array of optimized parameters for the model

        graphs:
        Figure 1  shows a scatter plot of your original geodata
        Figure 2  shows the experimental and theoretical semivariogram
        Figure 3  shows the number of samples in each lag and angle
        Figure 4  visualize di and td, the angles and angle tolerances
                  you have chosen.


        References
        ----------
        Small code parts based on Alghalandis.com.


        Examples
        --------
        # provide you some sample data:
        >>> # seed for reproducible results in doctest
        >>> np.random.seed(1)

        # easting
        >>> x = np.array([557509.27,557518.11,557526.95,557535.79,557544.63,
        ...               557553.47,557544.63,557535.79,557526.95,557518.11,
        ...               557526.95,557535.79,557544.63,557553.47,557562.31,
        ...               557571.15,557562.31,557553.47,557544.63,557535.79,
        ...               557544.63,557553.47,557562.31,557571.15,557579.99,
        ...               557597.66,557579.99,557562.31,557544.63,557526.7,
        ...               557509.27,557491.60,557473.92,557491.6,557509.27,
        ...               557526.95,557544.63,557562.31,557579.99,557597.66,
        ...               557615.34,557650.7,557686.05,557756.76,557827.47,
        ...               557898.18,557827.47,557756.76,557686.05,557615.34,
        ...               557650.70,557686.07,557721.41,557756.76,557721.41,
        ...               557686.05,557650.70,557686.05,557650.7,557579.99,
        ...               557615.34,557509.27,557579.99,557544.63,557509.27,
        ...               557473.92,557438.56,557403.21,557438.56,557473.92,
        ...               557509.27,557544.63,557579.99,557615.34,557615.34,
        ...               557579.99,557544.63,557509.27,557473.92,557438.56,
        ...               557403.21,557367.85,557332.5,557367.85,557403.21,
        ...               557438.56,557473.92,557332.50,557261.79,557191.08,
        ...               557261.79,557332.50,557403.21,557473.92,557544.63,
        ...               557615.34])

        # northing
        >>> y = np.array([4332422.55,4332413.71,4332404.87,4332396.03,
        ...               4332387.19,4332396.03,4332404.87,4332413.71,
        ...               4332422.55,4332431.39,4332440.23,4332431.39,
        ...               4332422.55,4332413.71,4332404.87,4332413.71,
        ...               4332422.55,4332431.39,4332440.23,4332449.07,
        ...               4332457.91,4332449.07,4332440.23,4332431.39,
        ...               4332422.55,4332404.87,4332387.19,4332369.52,
        ...               4332351.84,4332369.52,4332387.19,4332404.87,
        ...               4332422.55,4332440.23,4332457.91,4332475.58,
        ...               4332493.26,4332475.58,4332457.91,4332440.23,
        ...               4332422.55,4332316.48,4332210.42,4332281.13,
        ...               4332351.84,4332422.55,4332493.26,4332563.97,
        ...               4332634.68,4332563.97,4332528.62,4332493.25,
        ...               4332457.91,4332422.55,4332387.19,4332351.84,
        ...               4332387.19,4332422.55,4332457.91,4332599.33,
        ...               4332493.26,4332599.33,4332528.62,4332563.97,
        ...               4332528.62,4332493.26,4332457.91,4332422.55,
        ...               4332387.19,4332351.84,4332316.48,4332281.13,
        ...               4332316.48,4332351.84,4332281.13,4332245.77,
        ...               4332210.42,4332245.77,4332281.13,4332316.48,
        ...               4332351.84,4332387.19,4332422.55,4332457.91,
        ...               4332493.26,4332528.62,4332563.97,4332563.97,
        ...               4332493.26,4332422.55,4332351.84,4332281.13,
        ...               4332210.42,4332139.71,4332069.00,4332139.71])

        # value
        >>> v = np.array([9.94691161e-01,7.94158417e-02,0.00000000e+00,
        ...               1.75837990e+00,0.00000000e+00,0.00000000e+00,
        ...               0.00000000e+00,0.00000000e+00,1.02915310e+00,
        ...               2.69597379e+00,2.14552427e+00,2.18417112e+00,
        ...               0.00000000e+00,8.96101277e-01,1.14034753e+00,
        ...               3.46398689e-01,3.01418491e-01,0.00000000e+00,
        ...               1.17920343e+00,1.09682206e+00,4.79485665e-01,
        ...               0.00000000e+00,1.83183398e+00,0.00000000e+00,
        ...               0.00000000e+00,0.00000000e+00,9.86233407e-02,
        ...               7.68290376e-02,2.63911513e-01,0.00000000e+00,
        ...               2.10013460e+00,0.00000000e+00,2.47535521e+00,
        ...               1.47047869e+00,8.00371532e-01,2.39448347e+00,
        ...               0.00000000e+00,2.26426861e+00,0.00000000e+00,
        ...               0.00000000e+00,1.13769438e+00,1.01969271e+00,
        ...               2.26036007e+00,2.38991410e+00,1.82558084e-03,
        ...               0.00000000e+00,0.00000000e+00,2.52583544e+00,
        ...               6.35195403e-01,2.43778382e+00,0.00000000e+00,
        ...               2.47738704e+00,8.83280548e-01,2.42328547e+00,
        ...               0.00000000e+00,2.41534081e+00,2.45629467e+00,
        ...               0.00000000e+00,2.50770630e+00,1.30382267e+00,
        ...               2.06891940e+00,9.17384801e-02,0.00000000e+00,
        ...               1.10185544e-01,2.53460688e+00,2.15217780e+00,
        ...               1.16908154e+00,1.70072787e-01,1.60603658e-01,
        ...               2.15438377e+00,2.32464926e+00,3.26255002e-01,
        ...               0.00000000e+00,1.48404530e+00,2.10439439e+00,
        ...               0.00000000e+00,0.00000000e+00,0.00000000e+00,
        ...               2.34663663e-01,1.46993948e+00,2.67691613e+00,
        ...               2.13262460e-02,1.01551520e+00,1.10878523e+00,
        ...               1.80374874e+00,1.85571813e+00,2.93929948e+00,
        ...               4.43192829e-01,2.55962879e+00,0.00000000e+00,
        ...               1.46545683e+00,1.75659977e+00,0.00000000e+00,
        ...               2.37093751e+00,0.00000000e+00,0.00000000e+00])

        # omnidirectional semivariogram with exponential model and fifty lags
        >>> td = 180
        >>> di = [0]
        >>> nL = 50
        >>> h, g, c = semivariogram(x,y,v,nL,di,td,stype='omnidirectional',
        ...                         negscat=0., model='nomodel',graph=False,lunit='m',
        ...                         p0=(0.5,0.5,100),runtimediag=False)
        >>> from autostring import astr
        >>> print(astr(g[0][0:9], 3))
        ['0.550' '0.776' '0.619' '0.849' '0.991' '1.033' '1.067' '1.106' '1.079']
        >>> nugget,sill,irange,cork,h,g,c,semi_mod,semi_popt = semivariogram(
        ...      x,y,v,nL,di,td,stype='omnidirectional',
        ...      negscat=0.,model='exponential',graph=False,lunit='m',
        ...      p0=(0.5,0.5,1./100.), runtimediag=False)
        >>> print(astr(nugget, 1))
        ['0.5']
        >>> print(astr(sill, 1))
        ['1.1']
        >>> print(np.around(irange, -1))
        [80.]

        # directional semivariogram with spherical model and fifty lags
        >>> td = 45
        >>> di = [0,90]
        >>> nL = 50
        >>> nugget,sill,irange,cork,h,g,c,semi_mod,semi_popt = semivariogram(
        ...     x,y,v,nL,di,td,stype='directional',negscat=0.,
        ...     model='spherical',graph=False,lunit='m',p0=(0.5,0.5,100),
        ...     runtimediag=False)
        >>> print(astr(nugget, 1))
        ['0.5' '0.5']
        >>> print(astr(sill, 1))
        ['1.1' '1.1']

        # directional+orientational semivariogram with gaussian model and
        # fifty lags
        >>> td = 30
        >>> di = [0,45,90,135,180,-45,-90,-135]
        >>> nL = 50
        >>> nugget, sill, irange, cork, h, g, c, semi_mod, semi_popt = semivariogram(
        ...     x, y, v, nL, di, td, stype='directional+orientational',
        ...     negscat=0., model='spherical', graph=False, lunit='m',
        ...     p0=(0.5,0.5,100), runtimediag=False)
        >>> soll = [0.36, 0.42, 0.54, 0.50, 0.20, 0.57, 0.65, 0.45]
        >>> print(np.allclose(nugget, soll, atol=0.1))
        True
        >>> print(astr(sill, 1))
        ['1.0' '1.2' '1.1' '1.2' '1.0' '1.1' '1.1' '1.1']

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2011-2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Feb 2011
        Modified, AP, Nov 2012 - Parameter bounds for fitting algr.
                               - Minor errors corrected
                  MC, Nov 2012 - default graph=False
                  AP, Dec 2012 - documentation change
                  MC, Feb 2013 - change from fmin_l_bgfs_n to fmin_tnc because the former printed
                                 "Positive dir derivative in projection"
                                 "Using the backtracking step"
                                 even with iprint=-1, disp=0.
                  MC, Feb 2013 - ported to Python 3
                  MC, Apr 2014 - assert
    """

    # check input data
    assert (np.shape(x) == np.shape(y)) & (np.shape(y) == np.shape(v)), 'SemivariogramError: x, y and v must have the same dimensions.'

    di = np.array(di)
    if (stype == 'directional') & np.any(di<0.):
        raise TypeError('SemivariogramError: if you choose stype=directional,'
                        ' no negative directions in di are allowed')

    assert not (np.any(di>180.) | np.any(di<(-180.))), 'SemivariogramError: elements of di > 180 or < -180 are not allowed.'

    assert td <= 180, 'SemivariogramError: td > 180 is not allowed.'

    if model != 'nomodel':
        try:
            # from scipy.optimize import fmin_l_bfgs_b
            from scipy.optimize import fmin_tnc
        except ImportError:
            model = 'nomodel'
            print('SemivariogramWarning: the package scipy.optimize.curve_fit can'
                  ' not be found. It requires at least scipy 0.8.0. No fitting of'
                  ' experimental semivariogram(s) is possible.')

    if runtimediag:
        import time
        start = time.time()

    #---------------------------------------
    #compute all distances and angles between vectors x and y
    r, t, xr, n = distang(x,y)

    t = np.where(t==-180, 180, t)

    if stype == 'omnidirectional':
        headtitle = 'Omnidirectional Semivariogram'
        di = np.array([0])
        td = 180
    elif stype == 'directional':
        headtitle = 'Directional Semivariograms'
    elif stype == 'directional+orientational':
        headtitle = 'Directional and Orientational Semivariograms'
    else:
        raise NameError('SemivariogramError: wrong semivariogram type! '
         'Choose omnidirectional, directional or directional+orientational')

    #---------------------------------------
    # compute semivariogram
    h = []
    g = []
    c = []
    for i in di:
        hi, gi, ci = semivario(r, t, xr, n, v, nL, i, td, stype=stype, negscat=negscat)
        h.append(hi)
        g.append(gi)
        c.append(ci)

    if runtimediag:
        stop1 = time.time()
        print('Experimental semivariogram took %0.3f seconds' %(stop1-start))

    if model == 'exponential':
        func  = [expvar]
        irange = [exprange]
        subtitles = ['Exponential Semivariogram Model']
    elif model == 'spherical':
        func  = [sphvar]
        irange = [sphrange]
        subtitles = ['Spherical Semivariogram Model']
    elif model == 'gaussian':
        func  = [gauvar]
        irange = [gaurange]
        subtitles = ['Gaussian Semivariogram Model' ]
    elif model == 'noidea':
        func  = [expvar, sphvar, gauvar]
        irange = [exprange,sphrange,gaurange]
        subtitles = ['Exponential Semivariogram Model',
                     'Spherical Semivariogram Model',
                     'Gaussian Semivariogram Model' ]
    elif model == 'nomodel':
        subtitles = ['No Semivariogram Model']
    else:
        raise NameError('SemivariogramError: wrong semivariogram model! '
         'Choose exponential, spherical or gaussian')

    #---------------------------------------
    # fit semivariogram model
    if model != 'nomodel':
        popt = np.zeros((len(func),di.size,3))
        cork = np.zeros((len(func),di.size))
        for j in range(len(func)):
            for i in range(di.size):
                try:
                    # popti, f, d = fmin_l_bfgs_b(absdif, p0,
                    #                             args=(h[i], g[i], func[j]),
                    #                             approx_grad=1,
                    #                             bounds=[(0.,None),(0.,None),
                    #                                     (0.,None)],
                    #                             iprint=-1, disp=0)
                    popti, f, d = fmin_tnc(absdif, p0,
                                                args=(h[i], g[i], func[j]),
                                                approx_grad=1,
                                                bounds=[(0.,None),(0.,None),
                                                        (0.,None)],
                                                disp=0)
                    popt[j,i,:] = popti
                    cork[j,i] = np.corrcoef(g[i], func[j](h[i],popti))[0,1]
                except AttributeError:
                    print(('SemivariogramWarning: %s can not be fitted to' +
                           ' %i deg angle') %(subtitles[j], di[i]))

        ind = np.argmax(np.mean(cork, axis=1))

        if runtimediag:
            stop2 = time.time()
            print('Theoretical semivariogram took %0.3f seconds' %(stop2-stop1))

        nugget  = popt[ind,:,0]
        sill    = popt[ind,:,0] + popt[ind,:,1]
        irange  = irange[ind](sill,popt[ind,:,:])
        cork    = cork[ind]

        if runtimediag:
            print('Model: %s'%subtitles[ind])
            print('Nugget(s): %s'%nugget)
            print('Sill(s): %s'%sill)
            print('Range(s): %s'%irange)
            print('Coefficient(s) of correlation: %s'%cork)

    #---------------------------------------
    # plot
    if graph:
        import matplotlib.cm as cm
        import matplotlib.pyplot as plt
        import matplotlib as mpl

        mpl.rc('font', size=20)
        mpl.rc('lines', linewidth=2)
        mpl.rc('axes', linewidth=1.5)
        # mpl.rc('xtick.major', width=1.5)
        # mpl.rc('ytick.major', width=1.5)
        mpl.rcParams['lines.markersize']=6
        mpl.rcParams['lines.markeredgewidth']=1
        mpl.rcParams['grid.linewidth']=1.5
        # mpl.rcParams['legend.frameon']=False
        mpl.rcParams['legend.numpoints']=1
        mpl.rcParams['legend.handlelength']=1
        mpl.rcParams['mathtext.default']='regular'

        # scatterplot of input data
        # fig1 = plt.figure('scatter plot')
        fig1 = plt.figure(1)
        sub1 = fig1.add_subplot(111, aspect='equal')
        scat = sub1.scatter(x,y,c=v,s=40,cmap=plt.cm.jet)
        sub1.xaxis.set_major_locator(mpl.ticker.MultipleLocator(10))
        sub1.yaxis.set_major_locator(mpl.ticker.MultipleLocator(10))
        sub1.xaxis.set_major_formatter(mpl.ticker.
                                       ScalarFormatter(useOffset=False))
        sub1.yaxis.set_major_formatter(mpl.ticker.
                                       ScalarFormatter(useOffset=False))
        sub1.grid('on')
        sub1.set_title('scatter plot')
        plt.xlabel('easting [%s]' %lunit)
        plt.ylabel('northing [%s]' %lunit)
        fig1.autofmt_xdate(rotation=45)
        # plt.tight_layout(pad=1, h_pad=0, w_pad=0)
        # cbar need to be below autofm_xdate !!!???
        cbar = fig1.colorbar(scat,orientation='vertical',pad=0.05,shrink = 0.7)
        cbar.set_label('value')

        # semivariogram plot with/without fitted model
        # fig2 = plt.figure('semivariogram')
        fig2 = plt.figure(2)
        sub2 = fig2.add_subplot(111)
        sub2.plot(0,0,'.',c='k',markersize=1)
        for i in range(di.size):
            sub2.plot(h[i], g[i], 'o--', c=cm.jet(256/di.size*i),
                      label='%i'%di[i])
            if model != 'nomodel':
                h_mod_step = (np.max(h[i])-np.min(h[i]))/100.
                h_mod = np.arange(0.,np.max(h[i])+h_mod_step,h_mod_step)
                sub2.plot(h_mod, func[ind](h_mod, popt[ind,i,:]), '-',
                          c=cm.jet(256/di.size*i))
                sub2.plot([irange[i],irange[i]],
                          [0.,func[ind](irange[i], popt[ind,i,:])], '-.',
                          c=cm.jet(256/di.size*i))
                sub2.plot(irange[i],func[ind](irange[i], popt[ind,i,:]), 'rs')
        plt.xlabel('h [%s]' %lunit)
        plt.ylabel('$\\gamma(h)$')
        if model != 'nomodel':
            sub2.set_title('%s \n %s' %(headtitle, subtitles[ind]))
        else:
            sub2.set_title('%s' %(headtitle))
        plt.axis('auto')
        plt.grid('on')
        plt.legend(loc='best')
        # plt.tight_layout(pad=1, h_pad=0, w_pad=0)

        # plot samples per lag
        # fig3 = plt.figure('samples per lag')
        fig3 = plt.figure(3)
        sub3 = fig3.add_subplot(111)
        for i in range(di.size):
            sub3.plot(h[i],c[i],'o--', c=cm.jet(256/di.size*i),
                      label='%i'%di[i])
        plt.xlabel('h [%s]' %lunit)
        plt.ylabel('Samples per lag')
        plt.title('%s' %(headtitle))
        plt.axis('auto')
        plt.grid('on')
        plt.legend(loc='best')
        # plt.tight_layout(pad=1, h_pad=0, w_pad=0)

        # plot of directions
        # fig4 = plt.figure('directions', figsize=(6,6))
        fig4 = plt.figure(4, figsize=(6,6))
        ax = fig4.add_axes([0.1, 0.1, 0.8, 0.8], polar=True, axisbg='black')
        if stype == 'directional':
            theta = np.deg2rad(np.append((di-td),(di-td)+180))
        else:
            theta = np.deg2rad(di-td)
        radii = np.ones_like(theta)
        radii[::2] += 0.1
        width = np.deg2rad(np.ones_like(theta)*td*2)
        xlab =[]
        for i in range(19):
            #xlab.append(r'$\sf{%i\degree}$' %(i*10))
            xlab.append('$%i\\degree$' %(i*10))
        for i in range(17):
            #xlab.append(r'$\sf{%i\degree}$' %(-170+i*10))
            xlab.append('$%i\\degree$' %(-170+i*10))
        bars = ax.bar(theta, radii, width=width, bottom=0.0)

        for r,bar in zip(radii, bars):
            bar.set_facecolor('#00ff04')
            bar.set_edgecolor('yellow')
            bar.set_alpha(0.5)
            ax.set_yticklabels(radii, alpha=0)
            ax.xaxis.set_major_locator(plt.LinearLocator(numticks=37,
                                                         presets=None))
            ax.set_xticklabels(xlab, fontsize=15)
            plt.grid(color='yellow')
        ax.set_title(headtitle, fontsize=15)
        plt.show()

    # output
    if model != 'nomodel':
        return nugget, sill, irange, cork, h, g, c, func[ind], popt[ind,:,:]
    else:
        return h, g, c

#---------------------------------------
# sub functions
#---------------------------------------
# function to compute all distances and angles between vectors x and y
def distang(x,y):
    n = x.size
    t = np.zeros(n*(n-1)//2)
    r = np.zeros(n*(n-1)//2)
    k = 0
    for o in range(n-1):
        for p in range(o+1,n):
            dx = x[p]-x[o]
            dy = y[p]-y[o]
            t[k] = np.arctan2(dy,dx)       # angle (theta)
            r[k] = np.sqrt(dx**2+dy**2)    # distance (ray)
            k += 1
    xr = max(r)
    return r, t, xr, n

#---------------------------------------
# function to compute semivariogram
def semivario(r, t, xr, n, z, nL, di, td, stype='omnidirectional', negscat=0.):
    L = xr/nL                           # lag size
    g = np.zeros(nL)
    c = np.zeros(nL)
    a = np.deg2rad(di)                     # angle = direction in radian
    ta = np.deg2rad(td)                    # tolerance of angle in radians
    b = np.deg2rad(0.01)                   # buffer for numerical problems
    for s in range(nL):
        k = 0
        q = 0
        g[s] = 0
        c[s] = 0
        for o in range(n-1):
            for p in range(o+1,n):
                if  (s*L < abs(r[k]) < (s+1)*L):
                    if stype == 'omnidirectional':
                        g[s] += (z[p]-z[o])**2
                        q += 1
                    if stype == 'directional':
                        if a<0:
                            m=np.deg2rad(di+180)
                        else:
                            m=np.deg2rad(di-180)

                        if a+ta+b > np.deg2rad(180):
                            if (((a-ta-b)<t[k]<a) or (a<t[k]<np.deg2rad(180)+b)
                                 or (-np.deg2rad(180)-b<t[k]<(-np.deg2rad(360)+
                                                              (a+ta+b)))):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                        elif a-ta-b < -np.deg2rad(180):
                            if (((a+ta+b)>t[k]>a) or (a>t[k]>-np.deg2rad(180)-b)
                                 or (np.deg2rad(180)+b>t[k]>(np.deg2rad(360)+
                                                             (a-ta-b)))):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                        else:
                            if (a-ta-b)<t[k]<(a+ta+b):
                                g[s] += (z[p]-z[o])**2
                                q += 1

                        if m+ta+b > np.deg2rad(180):
                            if (((m-ta-b)<t[k]<m) or (m<t[k]<np.deg2rad(180)+b)
                                 or (-np.deg2rad(180)-b<t[k]<(-np.deg2rad(360)+
                                                              (m+ta+b)))):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                        elif m-ta-b < -np.deg2rad(180):
                            if (((m+ta+b)>t[k]>m) or (m>t[k]>-np.deg2rad(180)-b)
                                 or (np.deg2rad(180)+b>t[k]>(np.deg2rad(360)+
                                                             (m-ta-b)))):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                        else:
                            if (m-ta-b)<t[k]<(m+ta+b):
                                g[s] += (z[p]-z[o])**2
                                q += 1

                    if stype == 'directional+orientational':
                        if a+ta+b > np.deg2rad(180):
                            if (((a-ta-b)<t[k]<a) or (a<t[k]<np.deg2rad(180)+b)
                                 or (-np.deg2rad(180)-b<t[k]<(-np.deg2rad(360)+
                                                              (a+ta+b)))):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                        elif a-ta-b < -np.deg2rad(180):
                            if (((a+ta+b)>t[k]>a) or (a>t[k]>-np.deg2rad(180)-b)
                                 or (np.deg2rad(180)+b>t[k]>(np.deg2rad(360)+
                                                             (a-ta-b)))):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                        else:
                            if (a-ta-b)<t[k]<(a+ta+b):
                                g[s] += (z[p]-z[o])**2
                                q += 1
                k += 1
        g[s] /= (q*2) if q>0 else np.NAN
        c[s] = q
    h = np.array(np.arange(nL))*L+L/2
    h = np.delete(h,np.where(np.isnan(g)))       # ranges
    c = np.delete(c,np.where(np.isnan(g)))       # number of samples per range
    g = np.delete(g,np.where(np.isnan(g)))       # variance per range

    if (negscat > 0.) and (negscat < 1.):
        h = h[0:int(np.size(h)*negscat)]
        c = c[0:int(np.size(c)*negscat)]
        g = g[0:int(np.size(g)*negscat)]

    return h,g,c

#---------------------------------------
# exponential semivariogram model
def expvar(h,p):
    # return p[0] + p[1]*(1.-np.exp(-abs(h)/p[2]))
    return p[0] + p[1]*(1.-np.exp(-abs(h)*p[2]))

# range of exponential model at a given sill
def exprange(sill,p):
    # return -np.log((-sill*0.95+p[:,0])/p[:,1]+1)*p[:,2]
    return -np.log((-sill*0.95+p[:,0])/p[:,1]+1)/p[:,2]

#---------------------------------------
# spherical semivariogram model
def sphvar(h,p):
    return np.where((0<=np.abs(h)) & (np.abs(h)<=p[2]) & (p[2]!=0.),
                   p[0]+p[1]*((3./2.)*(abs(h)/p[2])-(1./2.)*(abs(h)/p[2])**3),
                   p[0]+p[1])
    # out1 = p[0]+p[1]
    # out2 = out1
    # if p[2] != 0.: out2 = p[0]+p[1]*(1.5*(abs(h)/p[2]) - 0.5*(abs(h)/p[2])**3
    # return np.where((np.abs(h)>=0.) & (np.abs(h)<=p[2]), out1, out2)

# range of spherical model at a given sill
def sphrange(sill,p):
    return p[:,2]

#---------------------------------------
# gaussian semivariogram model
def gauvar(h,p):
    # return p[0] + p[1]*(1. - np.exp(-h**2/p[2]**2))
    return p[0] + p[1]*(1. - np.exp(-h**2 * p[2]**2))

# range of gaussian model at a given sill
def gaurange(sill,p):
    # return np.sqrt( -(p[:,2])**2 * np.log((-sill*0.95+p[:,0])/p[:,1]+1))
    return np.sqrt( -(1./p[:,2])**2 * np.log((-sill*0.95+p[:,0])/p[:,1]+1.))

#---------------------------------------
# objective function
def absdif(p,x,y,func):
        return np.sum(np.abs(y-func(x, p)))


# DOCTEST:
if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # # easting
    # x = np.array([557509.27,557518.11,557526.95,557535.79,557544.63,
    #               557553.47,557544.63,557535.79,557526.95,557518.11,
    #               557526.95,557535.79,557544.63,557553.47,557562.31,
    #               557571.15,557562.31,557553.47,557544.63,557535.79,
    #               557544.63,557553.47,557562.31,557571.15,557579.99,
    #               557597.66,557579.99,557562.31,557544.63,557526.7,
    #               557509.27,557491.60,557473.92,557491.6,557509.27,
    #               557526.95,557544.63,557562.31,557579.99,557597.66,
    #               557615.34,557650.7,557686.05,557756.76,557827.47,
    #               557898.18,557827.47,557756.76,557686.05,557615.34,
    #               557650.70,557686.07,557721.41,557756.76,557721.41,
    #               557686.05,557650.70,557686.05,557650.7,557579.99,
    #               557615.34,557509.27,557579.99,557544.63,557509.27,
    #               557473.92,557438.56,557403.21,557438.56,557473.92,
    #               557509.27,557544.63,557579.99,557615.34,557615.34,
    #               557579.99,557544.63,557509.27,557473.92,557438.56,
    #               557403.21,557367.85,557332.5,557367.85,557403.21,
    #               557438.56,557473.92,557332.50,557261.79,557191.08,
    #               557261.79,557332.50,557403.21,557473.92,557544.63,
    #               557615.34])

    # # northing
    # y = np.array([4332422.55,4332413.71,4332404.87,4332396.03,
    #               4332387.19,4332396.03,4332404.87,4332413.71,
    #               4332422.55,4332431.39,4332440.23,4332431.39,
    #               4332422.55,4332413.71,4332404.87,4332413.71,
    #               4332422.55,4332431.39,4332440.23,4332449.07,
    #               4332457.91,4332449.07,4332440.23,4332431.39,
    #               4332422.55,4332404.87,4332387.19,4332369.52,
    #               4332351.84,4332369.52,4332387.19,4332404.87,
    #               4332422.55,4332440.23,4332457.91,4332475.58,
    #               4332493.26,4332475.58,4332457.91,4332440.23,
    #               4332422.55,4332316.48,4332210.42,4332281.13,
    #               4332351.84,4332422.55,4332493.26,4332563.97,
    #               4332634.68,4332563.97,4332528.62,4332493.25,
    #               4332457.91,4332422.55,4332387.19,4332351.84,
    #               4332387.19,4332422.55,4332457.91,4332599.33,
    #               4332493.26,4332599.33,4332528.62,4332563.97,
    #               4332528.62,4332493.26,4332457.91,4332422.55,
    #               4332387.19,4332351.84,4332316.48,4332281.13,
    #               4332316.48,4332351.84,4332281.13,4332245.77,
    #               4332210.42,4332245.77,4332281.13,4332316.48,
    #               4332351.84,4332387.19,4332422.55,4332457.91,
    #               4332493.26,4332528.62,4332563.97,4332563.97,
    #               4332493.26,4332422.55,4332351.84,4332281.13,
    #               4332210.42,4332139.71,4332069.00,4332139.71])
    # # value
    # v = np.array([9.94691161e-01,7.94158417e-02,0.00000000e+00,
    #               1.75837990e+00,0.00000000e+00,0.00000000e+00,
    #               0.00000000e+00,0.00000000e+00,1.02915310e+00,
    #               2.69597379e+00,2.14552427e+00,2.18417112e+00,
    #               0.00000000e+00,8.96101277e-01,1.14034753e+00,
    #               3.46398689e-01,3.01418491e-01,0.00000000e+00,
    #               1.17920343e+00,1.09682206e+00,4.79485665e-01,
    #               0.00000000e+00,1.83183398e+00,0.00000000e+00,
    #               0.00000000e+00,0.00000000e+00,9.86233407e-02,
    #               7.68290376e-02,2.63911513e-01,0.00000000e+00,
    #               2.10013460e+00,0.00000000e+00,2.47535521e+00,
    #               1.47047869e+00,8.00371532e-01,2.39448347e+00,
    #               0.00000000e+00,2.26426861e+00,0.00000000e+00,
    #               0.00000000e+00,1.13769438e+00,1.01969271e+00,
    #               2.26036007e+00,2.38991410e+00,1.82558084e-03,
    #               0.00000000e+00,0.00000000e+00,2.52583544e+00,
    #               6.35195403e-01,2.43778382e+00,0.00000000e+00,
    #               2.47738704e+00,8.83280548e-01,2.42328547e+00,
    #               0.00000000e+00,2.41534081e+00,2.45629467e+00,
    #               0.00000000e+00,2.50770630e+00,1.30382267e+00,
    #               2.06891940e+00,9.17384801e-02,0.00000000e+00,
    #               1.10185544e-01,2.53460688e+00,2.15217780e+00,
    #               1.16908154e+00,1.70072787e-01,1.60603658e-01,
    #               2.15438377e+00,2.32464926e+00,3.26255002e-01,
    #               0.00000000e+00,1.48404530e+00,2.10439439e+00,
    #               0.00000000e+00,0.00000000e+00,0.00000000e+00,
    #               2.34663663e-01,1.46993948e+00,2.67691613e+00,
    #               2.13262460e-02,1.01551520e+00,1.10878523e+00,
    #               1.80374874e+00,1.85571813e+00,2.93929948e+00,
    #               4.43192829e-01,2.55962879e+00,0.00000000e+00,
    #               1.46545683e+00,1.75659977e+00,0.00000000e+00,
    #               2.37093751e+00,0.00000000e+00,0.00000000e+00])
    # # omnidirectional semivariogram with exponential model and fifty lags
    # td = 180
    # di = [0]
    # nL = 50
    # h, g, c = semivariogram(x, y, v, nL, di, td, stype='omnidirectional',
    #                         negscat=False, model='nomodel', graph=False, lunit='m',
    #                         p0=(0.5,0.5,100.), runtimediag=False)
    # print np.round(g,3)
    # # [[ 0.55   0.776  0.619  0.849  0.991  1.033  1.067  1.106  1.079  1.002
    # #    1.119  1.06   0.935  0.789  1.117  1.125  1.063  1.015  1.041  0.911
    # #    1.077  1.116  1.202  1.195  0.935  0.918  0.946  0.596  1.035  1.153
    # #    1.205  1.635  1.384  0.806  0.006  1.409  0.748  0.31   0.977  1.95
    # #    1.748  1.087]]
    # nugget, sill, irange, cork, h, g, c, semi_mod, semi_popt = semivariogram(
    #     x, y, v, nL, di, td, stype='omnidirectional',
    #     negscat=False, model='exponential', graph=False, lunit='m',
    #     p0=(0.5,0.5,1./100.), runtimediag=False)
    # print round(nugget, 2)
    # # 0.41
    # print round(sill, 2)
    # # 1.06
    # print round(irange, 0)
    # # 73.0

    # # directional semivariogram with spherical model and fifty lags
    # td = 45
    # di = [0, 90]
    # nL = 50
    # nugget, sill, irange, cork, h, g, c, semi_mod, semi_popt = semivariogram(
    #     x, y, v, nL, di, td, stype='directional',
    #     negscat=False, model='spherical', graph=False, lunit='m',
    #     p0=(0.5,0.5,100), runtimediag=False)
    # print np.round(nugget, 2)
    # # [ 0.49  0.5 ]
    # print np.round(sill, 2)
    # # [ 1.08  1.07]

    # # directional+orientational semivariogram with gaussian model and
    # # fifty lags
    # td = 30
    # di = [0, 45, 90, 135, 180, -45, -90, -135]
    # nL = 50
    # nugget, sill, irange, cork, h, g, c, semi_mod, semi_popt = semivariogram(
    #     x, y, v, nL, di, td, stype='directional+orientational',
    #     negscat=False, model='spherical', graph=True, lunit='m',
    #     p0=(0.5,0.5,100.), runtimediag=False)
    # print np.round(nugget, 2)
    # # [ 0.25  0.4   0.58  0.49  0.32  0.61  0.49  0.42]
    # print np.round(sill, 2)
    # # [ 0.98  1.12  1.15  1.18  1.01  1.11  0.99  1.09]




#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
from scipy.spatial.distance import cdist


__all__ = ['homo_sampling']


def homo_sampling(p1, p2, n, func=None, plot=False, maxit=1000):
    """
    Generates homogeneously, randomly distributed sample points within a
    given rectangular area spanned by p1 (lower left) and p2 (upper right)
    corner. The area can be masked by a user given function to exclude
    certain areas of the rectangle from sampling.


    Definition
    ----------
    homo_sampling(p1, p2, n, func=None, plot=False, maxit=1000):


    Input
    -----
    p1          np.array((1,2)), coordinates of lower left corner (x,y)
    p2          np.array((1,2)), coordinates of upper right corner (x,y)
    n           int, number of points to be generated


    Optional Input
    --------------
    func        callable function of the form func(xy) defined by user with
                xy beeing a np.array((m,2)), returning a boolean
                np.array((m)) with True when points in xy are NOT in the
                desired area and False when points are in the desired area
    plot        bool, if True results are plotted (default=False)
    maxit       int, maximum number of iterations for maximizing the minimum
                distance between points (default=1000)


    Output
    ------
    xy_out      np.array((n,2)), x and y coordinates of points


    Restrictions
    ------------
    Does not deal with masked arrays.
    TODO:
        - a priory estimation of masked area fraction to increase n
          internally to reduce necessary iterations
        - include running window of max(minimum distance between points)
          to stop iteration when no further increase occurs within
          window-width=maxit


    Examples
    --------
    >>> # Define the corner points of a rectangle
    >>> p1 = np.array([644242.398, 5766315.198])
    >>> p2 = np.array([644555.611, 5766578.860])

    >>> # Generate 10 random points within the rectangle
    >>> xy = homo_sampling(p1, p2, 10)

    >>> # Increase maximum number of iterations to distribute points
    >>> # even more homogeneous
    >>> xy = homo_sampling(p1, p2, 10, maxit=10000)

    >>> # Define a function that excludes the area outside a given radius r
    >>> # around a central point cp
    >>> def func(xy):
    ...     cp = np.array([[644394.412, 5766429.022]])
    ...     r = 100.
    ...     from scipy.spatial.distance import cdist
    ...     return cdist(cp, xy, metric='euclidean')[0] > r

    >>> # Generate points only within the radius around cp
    >>> xy = homo_sampling(p1, p2, 10, func=func)

    >>> # Since the result are random numbers and can not be restricted to
    >>> # a certain seed, no result comparison within doctest can be performed.


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2021 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Arndt Piayda,   Jul 2014
    Modified, Matthias Cuntz, Sep 2021 - code refactoring
    """
    # check
    if (len(p1) != 2) or (len(p2) != 2):
        raise ValueError('homo_sampling: p1 and p2 must be of length 2')
    elif np.any(p2-p1 < 0) or np.any(p2 == p1):
        raise ValueError('homo_sampling: you mixed up the corners')
    elif func:
        if not callable(func):
            raise ValueError('homo_sampling: func must be a callable function')
        if np.ndim(func(np.array([[1., 1.]]))) != 1:
            raise ValueError('homo_sampling: func must return 1D np.array')
        if func(np.array([[1., 1.]])).dtype != bool:
            raise ValueError('homo_sampling: func must return boolean np.array')

    # generate 1st random point set
    xy  = np.empty((n, 2))
    rem = np.ones((n), dtype=bool)
    while np.any(rem):
        xy[rem, 0] = (p2[0] - p1[0])*np.random.random(np.sum(rem)) + p1[0]
        xy[rem, 1] = (p2[1] - p1[1])*np.random.random(np.sum(rem)) + p1[1]

        rem = func(xy) if func else False

    # maximize the distance between points
    conv = []
    tempxy = []

    for i in range(maxit):
        d   = cdist(xy, xy, metric='euclidean')
        d   = np.ma.array(d, mask=(d == 0.))
        rep = np.argmin(np.ma.amin(d, axis=0))

        conv   += [np.ma.amin(d)]
        tempxy += [np.copy(xy)]

        rem = np.ones((1), dtype=bool)
        while np.any(rem):
            xnew = (p2[0] - p1[0])*np.random.random(1) + p1[0]
            ynew = (p2[1] - p1[1])*np.random.random(1) + p1[1]

            rem = func(np.array([[xnew[0], ynew[0]]])) if func else False

        xy[rep, :] = [xnew[0], ynew[0]]

    xy_out = tempxy[np.argmax(conv)]

    if plot:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        fig = plt.figure('homo_sampling: minimum distance=%f'%(np.amax(conv)))
        sub1 = fig.add_subplot(122, aspect='equal')
        sub1.set_xlim(p1[0], p2[0])
        sub1.set_ylim(p1[1], p2[1])
        sub1.scatter(xy_out[:, 0], xy_out[:, 1])
        sub1.set_xlabel('X')
        sub1.set_ylabel('Y')

        sub2 = fig.add_subplot(121)
        sub2.plot(range(maxit), conv)
        sub2.plot(np.argmax(conv), np.amax(conv), 'o')
        sub2.set_xlabel('iterations')
        sub2.set_ylabel('minimum distance between points')
        plt.show()

        print(np.argmax(conv))
        print(np.amax(conv))

    return xy_out


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
This is the unittest for the const module.

python -m unittest -v tests/test_const.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_const.py

"""
import unittest


class TestConst(unittest.TestCase):
    """
    Tests for const/const.py

    """

    def test_const(self):
        import numpy as np
        from pyjams.const import Pi, pi, Pi2, pi2, Pi3, pi3
        from pyjams.const import TwoPi, Twopi, Sqrt2, sqrt2
        from pyjams.const import gravity, T0, P0, T25
        from pyjams.const import sigma, R, Rair, Rh2o, Na, kB, REarth
        from pyjams.const import mmol_co2, mmol_h2o, mmol_air
        from pyjams.const import molmass_co2, molmass_h2o, molmass_air
        from pyjams.const import density_quartz, cheat_quartz, cheat_water
        from pyjams.const import cheat_air, latentheat_vaporization
        from pyjams.const import R13VPDB, R18VSMOW, R2VSMOW
        from pyjams.const import eps, huge, tiny

        self.assertEqual(np.around(Pi, 4), 3.1416)
        self.assertEqual(np.around(pi, 4), 3.1416)
        self.assertEqual(np.around(Pi2, 4), 1.5708)
        self.assertEqual(np.around(pi2, 4), 1.5708)
        self.assertEqual(np.around(Pi3, 4), 1.0472)
        self.assertEqual(np.around(pi3, 4), 1.0472)
        self.assertEqual(np.around(TwoPi, 4), 6.2832)
        self.assertEqual(np.around(Twopi, 4), 6.2832)
        self.assertEqual(np.around(Sqrt2, 4), 1.4142)
        self.assertEqual(np.around(sqrt2, 4), 1.4142)
        self.assertEqual(gravity, 9.80665)
        self.assertEqual(T0, 273.15)
        self.assertEqual(P0, 101325.)
        self.assertEqual(T25, 298.15)
        self.assertEqual(sigma, 5.67e-08)
        self.assertEqual(np.around(R, 4), 8.3145)
        self.assertEqual(np.around(Rair, 4), 287.058)
        self.assertEqual(np.around(Rh2o, 4), 461.5228)
        self.assertEqual(Na, 6.02214076e23)
        self.assertEqual(kB, 1.380649e-23)
        self.assertEqual(REarth, 6371009.)
        self.assertEqual(mmol_co2, 44.009)
        self.assertEqual(mmol_h2o, 18.01528)
        self.assertEqual(mmol_air, 28.9644)
        self.assertEqual(molmass_co2, 44.009e-3)
        self.assertEqual(molmass_h2o, 18.01528e-3)
        self.assertEqual(molmass_air, 28.9644e-3)
        self.assertEqual(density_quartz, 2.65)
        self.assertEqual(cheat_quartz, 800.)
        self.assertEqual(cheat_water, 4180.)
        self.assertEqual(cheat_air, 1010.)
        self.assertEqual(latentheat_vaporization, 2.45e6)
        self.assertEqual(R13VPDB, 0.0112372)
        self.assertEqual(R18VSMOW, 2005.2e-6)
        self.assertEqual(R2VSMOW, 155.76e-6)
        self.assertEqual(eps, np.finfo(float).eps)
        self.assertEqual(huge, np.finfo(float).max)
        self.assertEqual(tiny, np.finfo(float).tiny)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np


__all__ = ['autostring', 'astr']


def autostring(num, prec=0, zero=False, set_printoptions=False, pp=False, join=False, joinall=False, sep=' '):
    """
    Format number (array) with given decimal precision.


    Definition
    ----------
    def autostring(num, prec=0, zero=False, set_printoptions=False, pp=False,
                   join=False, joinall=False, sep=' '):

    There is a wrapper function for convenience with the short name 'astr'
    that calls autostring
    def astr(num, prec=0, zero=False, set_printoptions=False, pp=False,
             join=False, joinall=False, sep=' '):


    Input
    -----
    num                 number array


    Optional Input
    --------------
    prec
        number of decimal places of formatted values
        minimum field width for integers (default: 0)
    zero
        if True, pad values with zeros rather than blanks (default: False)
    set_printoptions
        if True, sets linewidth to the format times size of 1st dimension
        (default: False)
    pp
        shortcut for set_printoptions (default: False)
        it will be checked for (pp | set_printoptions)
    join
        if True, joins all individual strings of last (fastest) dimension into
        one string (default: False)
    joinall
        if True, joins all individual strings into single string,
        i.e. first flattens the array and then joins it
        (default: False, overwrites join)
    sep
        separator used when joining (default: space=' ')


    Output
    ------
    string (array) of formatted numbers


    Restrictions
    ------------
    None


    Examples
    --------
    >>> print(autostring(3.5967, 3))
    3.597

    >>> print(autostring(3.5967))
    4

    >>> print(autostring(3, 3))
      3

    >>> print(autostring(np.array([3.5967, 3.5964]), 3))
    ['3.597'  '3.596']

    >>> print(autostring(np.array([3.59, 1.123456e12]), 3))
    ['3.590e+00'  '1.123e+12']

    >>> print(autostring(np.array([3.59, 11.1234]), 3, zero=True))
    ['03.590'  '11.123']

    >>> print(autostring(np.array([3, 11])))
    [' 3' '11']

    >>> print(autostring(np.array([3, 11]), 3))
    ['  3' ' 11']

    >>> print(autostring(np.zeros((2,2), dtype=float), 1))
    [['0.0' '0.0']
     ['0.0' '0.0']]

    >>> np.set_printoptions(threshold=10)
    >>> print(autostring(np.zeros((2,10), dtype=float)[0:2,0:2], 1))
    [['0.0' '0.0']
     ['0.0' '0.0']]

    >>> print(autostring(np.zeros((2,10), dtype=float), 1, set_printoptions=True))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(autostring(np.zeros((2,10), dtype=float), 1, pp=True))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(autostring(np.zeros((2,10), dtype=float), 1, set_printoptions=False, pp=True))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(autostring(np.array([3.5967, 3.5964]), 3, join=True))
    3.597 3.596

    >>> print(autostring(np.zeros((2,10), dtype=float), 1, join=True, sep=';'))
    ['0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0'
     '0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0']

    >>> print(autostring(np.reshape(np.arange(20,dtype=float),(2,10)), 1, joinall=True, sep=';'))
     0.0; 1.0; 2.0; 3.0; 4.0; 5.0; 6.0; 7.0; 8.0; 9.0;10.0;11.0;12.0;13.0;14.0;15.0;16.0;17.0;18.0;19.0

    >>> print(autostring(np.reshape(np.arange(20,dtype=float),(2,10)), 1, joinall=True, sep=';'))
     0.0; 1.0; 2.0; 3.0; 4.0; 5.0; 6.0; 7.0; 8.0; 9.0;10.0;11.0;12.0;13.0;14.0;15.0;16.0;17.0;18.0;19.0

    >>> print(autostring(np.array([3, 11, np.inf])))
    ['  3' ' 11' 'inf']

    >>> print(autostring(np.array([3, 11, np.nan])))
    ['  3' ' 11' 'nan']

    >>> print(autostring(np.ma.array([3, 11, np.nan], mask=[False,True,False])))
    ['  3' '-- ' 'nan']

    >>> print(autostring(np.ma.array([3, 11, np.nan], mask=[False,False,True])))
    [' 3' '11' '--']


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2011-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Nov 2011 - from autostring.pro
    Modified, Matthias Cuntz, May 2012 - pp
              Matthias Cuntz, Dec 2012 - special treatment of -0.0 on output
              Matthias Cuntz, Feb 2013 - nan, inf and masked arrays
              Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Oct 2014 - isinstance
              Matthias Cuntz, Dec 2014 - tuple input
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # Check input
    if isinstance(num, (list, tuple)):
        num = np.array(num)
    isarr = np.ndim(num)
    if (isarr > 2):
        print("AUTOSTRING WARNING: autostring only works with scalars, 1D- and 2D arrays: return original array.")
        return num
    # Only treat int and float
    if (isarr == 0):
        try:
            typ = num.dtype
        except AttributeError:
            if (type(num) == float):
                typ = np.float64
            elif (type(num) == int):
                typ = np.int32
            else:
                typ = type(num)
    else:
        typ = num.dtype
    try:
        lfloat = np.float128     # Mac/*nix
    except AttributeError:
        try:
            lfloat = np.float96  # Windows
        except AttributeError:
            lfloat = np.float64
    if np.__version__ >= "1.6":
        if (typ in [np.float16, np.float32, np.float64, lfloat]):
            isfloat = True
        elif (typ in [np.int8, np.int16, np.int32, np.int64, np.uint8,
                      np.uint16, np.uint32, np.uint64]):
            isfloat = False
        else:
            print("AUTOSTRING WARNING: autostring cannot work with input type: return original array.")
            return num
    else:
        if (typ in [np.float32, np.float64, lfloat]):
            isfloat = True
        elif (typ in [np.int8, np.int16, np.int32, np.int64, np.uint8,
                      np.uint16, np.uint32, np.uint64]):
            isfloat = False
        else:
            print("AUTOSTRING WARNING: autostring cannot work with input type: return original array.")
            return num
    # Scalar to array if necessary; Special treatment of -0.0
    if (isarr == 0):
        if (num == 0):
            num = np.abs(num)
    else:
        if isinstance(num, np.ma.masked_array):
            num = np.ma.where(num == 0, 0, num)
        else:
            num = np.where(num == 0, 0, num)
    # Zero padding
    if zero:
        nix = '0'
    else:
        nix = ''
    #
    # If we deal with an array of numbers we take the largest for the format
    # deal with inf and nan
    hasmask = False
    hasnan  = False
    if (isarr == 0):
        if np.isnan(num):
            return 'nan'
        if np.isinf(num):
            return 'inf'
        abs_num = np.ma.abs(num)
        # leave room for the decimal point and the negative sign, if any
        if (num < 0.):
            num_sign_chars = 1
        else:
            num_sign_chars = 0
    else:
        if isinstance(num, np.ma.masked_array):
            if np.sum(num.mask) > 0:
                hasmask = True
            if num.count() > np.ma.sum(np.isfinite(num)):
                hasnan = True
        else:
            if num.size > np.sum(np.isfinite(num)):
                hasnan = True
        inum   = np.ma.array(num, mask=~np.isfinite(num), keep_mask=True)
        abs_num = np.ma.max(np.ma.abs(inum))
        # leave room for the decimal point and the negative sign, if any
        if (np.ma.min(inum) < 0.):
            num_sign_chars = 1
        else:
            num_sign_chars = 0
    #
    # Floating point
    if isfloat:  # number is a float, more or less
        if abs_num >= 1.e6:
            num_prefix_chars  = 1
            num_sci_not_chars = 4
            format_type       = 'e'
        elif ((abs_num < 1.e6) & (abs_num >= 1.)):
            nprefix = np.int_(np.log10(np.int32(abs_num)))+1
            # special treatment: the output prefix digits could
            # be one digit longer as the input prefix digits:
            # e.g. 99.99 => 100.0
            val               = np.around(abs_num*(10.**prec))/(10.**prec)
            nprefixval        = np.int_(np.log10(val))+1
            nprefix           = np.amax(np.array([nprefix, nprefixval],
                                                 dtype=int))
            num_prefix_chars  = nprefix
            num_sci_not_chars = 0
            format_type       = 'f'
        elif ((abs_num < 1.) & (abs_num >= 1.e-3)):
            num_prefix_chars  = 1
            num_sci_not_chars = 0
            format_type       = 'f'
        elif (abs_num == 0):
            num_prefix_chars  = 1
            num_sci_not_chars = 0
            format_type       = 'f'
        else:
            num_prefix_chars  = 1
            num_sci_not_chars = 4
            format_type       = 'e'
        #
        num_postfix_chars = prec
        num_total_chars   = (num_sign_chars + num_prefix_chars + 1 +
                             num_postfix_chars + num_sci_not_chars)
        if (prec == 0):  # no dot if prec=0
            num_total_chars -= 1
        if hasmask:  # need space for --
            if num_total_chars < 2:
                num_total_chars = 2
        if hasnan:  # need space for nan or inf
            if num_total_chars < 3:
                num_total_chars = 3
        format_string = ("{0:s}{1:s}{2:d}{3:s}{4:d}{5:s}{6:s}".format(
            '{0:', nix, num_total_chars, '.', num_postfix_chars, format_type,
            '}'))
    else:  # number is an integer
        format_type = 'd'
        if abs_num != 0:
            num_digits = np.int_(np.log10(abs_num))+1
        else:
            num_digits = 1
        num_total_chars = np.maximum(num_digits + num_sign_chars, prec)
        if hasmask:  # need space for --
            if num_total_chars < 2:
                num_total_chars = 2
        if hasnan:  # need space for nan or inf
            if num_total_chars < 3:
                num_total_chars = 3
        format_string = ("{0:s}{1:s}{2:d}{3:s}{4:s}".format(
            '{0:', nix, num_total_chars, format_type, '}'))
    #
    if (isarr == 0):
        out = format_string.format(num)
        # Special treatment of -0.0
        if float(out) == 0:
            out = format_string.format(0)
    else:
        fnum = num.flatten()
        nnum = fnum.size
        import sys
        if sys.hexversion > int('0x3000000', base=16):
            styp = 'U{0:d}'.format(num_total_chars)
        else:
            styp = 'S{0:d}'.format(num_total_chars)
        out = np.empty(nnum, dtype=styp)
        for i in range(nnum):
            if str(fnum[i]) == '--':
                sformat_string = ("{0:s}{1:d}s{2:s}".format(
                    '{0:', num_total_chars, '}'))
                out[i] = sformat_string.format('--')
            else:
                out[i] = format_string.format(fnum[i])
                if float(out[i]) == 0:
                    out[i] = format_string.format(0)
        out = np.reshape(out, num.shape)
        if (set_printoptions | pp):
            # num_total_chars+3 for '' and space, +isarr for []
            np.set_printoptions(
                linewidth=num.shape[-1]*(num_total_chars+3)+isarr,
                threshold=nnum+1)
        if (join | joinall):  # There should be reduction routines in numpy
            if ((isarr == 1) | ((isarr == 2) & joinall)):
                if (isarr == 2):
                    out = out.flatten()
                for i in range(out.size):
                    if (i == 0):
                        outc = out[i]
                    else:
                        outc = outc+sep+out[i]
            else:
                if sys.hexversion > int('0x3000000', base=16):
                    sform = 'U{0:d}'.format(
                        (len(out[0, 0])+len(sep))*out.shape[1])
                else:
                    sform = 'S{0:d}'.format(
                        (len(out[0, 0])+len(sep))*out.shape[1])
                outc = np.zeros(out.shape[0], dtype=sform)
                for j in range(out.shape[0]):
                    for i in range(out.shape[1]):
                        if (i == 0):
                            outc[j] = out[j, i]
                        else:
                            outc[j] = outc[j]+sep+out[j, i]
            out = outc

    # return formatted string
    return out


def astr(num, prec=0, zero=False, set_printoptions=False, pp=True, join=False,
         joinall=False, sep=' '):
    """
    Wrapper function for autostring with pp=True by default.
    def autostring(num, prec=0, zero=False, set_printoptions=False, pp=False,
                   join=False, joinall=False, sep=' '):


    Examples
    --------
    >>> print(astr(3.5967, 3))
    3.597

    >>> print(astr(3.5967))
    4

    >>> print(astr(3, 3))
      3

    >>> print(astr(np.array([3.5967, 3.5964]), 3))
    ['3.597' '3.596']

    >>> print(astr(np.array([3.59, 1.123456e12]), 3))
    ['3.590e+00' '1.123e+12']

    >>> print(astr(np.array([3.59, 11.1234]), 3, zero=True))
    ['03.590' '11.123']

    >>> print(astr(np.array([3, 11])))
    [' 3' '11']

    >>> print(astr(np.array([3, 11]), 3))
    ['  3' ' 11']

    >>> print(astr(np.zeros((2,2), dtype=float), 1))
    [['0.0' '0.0']
     ['0.0' '0.0']]

    >>> np.set_printoptions(threshold=10)
    >>> print(astr(np.zeros((2,10), dtype=float), 1))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(astr(np.zeros((2,10), dtype=float), 1, set_printoptions=True))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(astr(np.zeros((2,10), dtype=float), 1, pp=True))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(astr(np.zeros((2,10), dtype=float), 1, set_printoptions=False, pp=True))
    [['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']
     ['0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0' '0.0']]

    >>> print(astr(np.array([3.5967, 3.5964]), 3, join=True))
    3.597 3.596

    >>> print(astr(np.zeros((2,10), dtype=float), 1, join=True, sep=';'))
    ['0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0'
     '0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0;0.0']

    >>> print(astr(np.reshape(np.arange(20,dtype=float),(2,10)), 1, joinall=True, sep=';'))
     0.0; 1.0; 2.0; 3.0; 4.0; 5.0; 6.0; 7.0; 8.0; 9.0;10.0;11.0;12.0;13.0;14.0;15.0;16.0;17.0;18.0;19.0

    >>> print(astr(np.array([3, 11, np.inf])))
    ['  3' ' 11' 'inf']

    >>> print(astr(np.array([3, 11, np.nan])))
    ['  3' ' 11' 'nan']

    >>> print(astr(np.ma.array([3, 11, np.nan], mask=[False,True,False])))
    ['  3' '-- ' 'nan']

    >>> print(astr(np.ma.array([3, 11, np.nan], mask=[False,False,True])))
    [' 3' '11' '--']
    """
    return autostring(num, prec=prec, zero=zero,
                      set_printoptions=set_printoptions,
                      pp=pp, join=join, joinall=joinall, sep=sep)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # print(autostring(np.array([3, 11, np.nan])))
    # #['  3' ' 11' 'nan']

    # print(autostring(np.ma.array([3, 11, np.nan], mask=[False,True,False])))
    # #['  3' '-- ' 'nan']

    # print(autostring(np.ma.array([3, 11, np.nan], mask=[False,False,True])))
    # #['  3' ' 11' '-- ']



#!/usr/bin/env python
"""
This is the unittest for sce module.

python -m unittest -v tests/test_sce.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_sce.py

"""
import unittest
import os
import numpy as np
from scipy.optimize._constraints import Bounds
from scipy.optimize import rosen

from pytest import raises as assert_raises, warns
from numpy.testing import assert_equal, assert_almost_equal

from pyjams.sce import SCESolver, _strtobool
from pyjams import sce


def rosenbrock_args(x, a, b=100):
    """
    Rosenbrock, global optimum: f(a, a^2) = 0.0.

    """
    f = (a - x[0])**2 + b * (x[1] - x[0]**2)**2

    return f


def rosenbrock_kwargs(x, a=1, b=100):
    """
    Rosenbrock, global optimum: f(a, a^2) = 0.0.

    """
    f = (a - x[0])**2 + b * (x[1] - x[0]**2)**2

    return f


def deb03(x):
    r"""
    Deb 3 objective function from scipy.benchmarks

    This is a multimodal minimization problem defined as follows:

    .. math::

        f_{\text{Deb02}}(x) = - \frac{1}{N} \sum_{i=1}^n \sin^6 \left[ 5 \pi
        \left ( x_i^{3/4} - 0.05 \right) \right ]


    Here, :math:`n` represents the number of dimensions and :math:`x_i \in
    [0, 1]` for :math:`i = 1, ..., n`.

    *Global optimum*: :math:`f(x) = 0.0`. The number of global minima is
    :math:`5^n` that are evenly spaced in the function landscape, where
    :math:`n` represents the dimension of the problem.

    .. [1] Jamil, M. & Yang, X.-S. A Literature Survey of Benchmark Functions
    For Global Optimization Problems Int. Journal of Mathematical Modelling
    and Numerical Optimisation, 2013, 4, 150-194.

    """
    return -(1.0 / len(x)) * np.sum(np.sin(5 * np.pi * (x**0.75 - 0.05))**6)


# Some lines are not covered:
#   120-121: second args is empty dict
#   919: if npt < 1
#   942: redo random number if exactly 0 in sampling='open'
#   957-960: if lb and ub <=0 with sampling='log'
class TestSCESolver(unittest.TestCase):

    def setUp(self):
        self.x0 = [0.5, 0.1]
        self.lb = [0., 0.]
        self.ub = np.array([2., 2.])
        self.rosenx = [1., 1.]

    def negative_rosen(self, x):
        return -rosen(x)

    def test_defaults(self):
        # test that defaults are set correctly
        solver = SCESolver(rosen, self.x0, self.lb, self.ub)
        assert_equal(solver.sampling, 'half-open')
        assert_equal(solver.maxn, 1000)
        assert_equal(solver.kstop, 10)
        assert_equal(solver.pcento, 0.0001)
        assert_equal(solver.ngs, 2)
        assert_equal(solver.npg, 2 * len(self.lb) + 1)
        assert_equal(solver.nps, len(self.lb) + 1)
        assert_equal(solver.nspl, 2 * len(self.lb) + 1)
        assert_equal(solver.mings, 2)
        assert_equal(solver.peps, 0.001)
        assert_equal(solver.alpha, 0.8)
        assert_equal(solver.beta, 0.45)
        assert_equal(solver.maxit, False)
        assert_equal(solver.printit, 2)
        assert_equal(solver.polish, True)
        assert_equal(solver.restartfile1, '')
        assert_equal(solver.restartfile2, '')

        solver = SCESolver(rosen, self.x0, self.lb, self.ub,
                           sampling='left-half-open',
                           maxn=100, kstop=1, pcento=0.001,
                           ngs=20, npg=10, nps=20, nspl=10, mings=20,
                           peps=0.01,
                           alpha=0.9, beta=0.55, maxit=True, printit=1,
                           polish=False,
                           restart=False, restartfile1='sce.restart.npz',
                           restartfile2='sce.restart.txt')
        assert_equal(solver.sampling, 'left-half-open')
        assert_equal(solver.maxn, 100)
        assert_equal(solver.kstop, 1)
        assert_equal(solver.pcento, 0.001)
        assert_equal(solver.ngs, 20)
        assert_equal(solver.npg, 10)
        assert_equal(solver.nps, 20)
        assert_equal(solver.nspl, 10)
        assert_equal(solver.mings, 20)
        assert_equal(solver.peps, 0.01)
        assert_equal(solver.alpha, 0.9)
        assert_equal(solver.beta, 0.55)
        assert_equal(solver.maxit, True)
        assert_equal(solver.printit, 1)
        assert_equal(solver.polish, False)
        assert_equal(solver.restartfile1, 'sce.restart.npz')
        assert_equal(solver.restartfile2, 'sce.restart.txt')

        solver = SCESolver(rosen, self.x0, self.lb, self.ub,
                           sampling='left-half-open',
                           maxn=100, kstop=1, pcento=0.001,
                           ngs=20, npg=10, nps=20, nspl=10, mings=20,
                           peps=0.01,
                           alpha=0.9, beta=0.55, maxit=True, printit=1,
                           polish=False,
                           restart=True, restartfile1='sce.restart.npz',
                           restartfile2='sce.restart.txt')
        assert_equal(solver.sampling, 'left-half-open')
        assert_equal(solver.maxn, 100)
        assert_equal(solver.kstop, 1)
        assert_equal(solver.pcento, 0.001)
        assert_equal(solver.ngs, 20)
        assert_equal(solver.npg, 10)
        assert_equal(solver.nps, 20)
        assert_equal(solver.nspl, 10)
        assert_equal(solver.mings, 20)
        assert_equal(solver.peps, 0.01)
        assert_equal(solver.alpha, 0.9)
        assert_equal(solver.beta, 0.55)
        assert_equal(solver.maxit, True)
        assert_equal(solver.printit, 1)
        assert_equal(solver.polish, False)
        assert_equal(solver.restartfile1, 'sce.restart.npz')
        assert_equal(solver.restartfile2, 'sce.restart.txt')

        toremove = [solver.restartfile1, solver.restartfile2]
        for ff in toremove:
            if os.path.exists(ff):
                os.remove(ff)

    def test_SCESolver(self):
        solver = SCESolver(rosen, self.x0, self.lb, self.ub, maxn=100,
                           polish=False)
        result = solver.solve()
        assert_equal(result.fun, rosen(result.x))

        solver = SCESolver(rosen, self.x0, self.lb, self.ub, maxn=100,
                           polish=True)
        result = solver.solve()
        assert_equal(result.fun, rosen(result.x))

    def test_sce(self):
        # standard
        result = sce(rosen, self.x0, self.lb, self.ub)
        assert_equal(result.fun, rosen(result.x))

        # bounds
        result = sce(rosen, self.x0, self.lb, self.ub)
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        result = sce(rosen, self.x0, list(zip(self.lb, self.ub)))
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        result = sce(rosen, self.x0, Bounds(self.lb, self.ub))
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        result = sce(rosen, self.x0, self.lb[0], self.ub[0])
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        result = sce(rosen, self.x0, self.lb[0:1], self.ub[0:1])
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        # degenerated bounds
        x0 = [0.999, 0.5, 0.1]
        lb = [2., 0., 0.]
        ub = [2., 2., 2.]
        rosenx = [x0[0], 1., 1.]
        result = sce(rosen, x0, lb, ub)
        assert_almost_equal(result.x, rosenx, decimal=2)
        x0 = [0.999, 0.5, 0.1]
        lb = [3., 0., 0.]
        ub = [2., 2., 2.]
        rosenx = [x0[0], 1., 1.]
        with warns(UserWarning):
            result = sce(rosen, x0, lb, ub)
        assert_almost_equal(result.x, rosenx, decimal=2)

        # seed
        result = sce(rosen, self.x0, self.lb, self.ub, polish=False, seed=1)
        result2 = sce(rosen, self.x0, self.lb, self.ub, polish=False, seed=1)
        assert_equal(result.x, result2.x)
        assert_equal(result.nfev, result2.nfev)

        # restart
        restartfile1 = 'sce.restart.npz'
        restartfile2 = restartfile1 + '.txt'
        result = sce(rosen, self.x0, self.lb, self.ub, polish=False, seed=1)
        _ = sce(rosen, self.x0, self.lb, self.ub, polish=False, seed=1,
                restart=False, restartfile1=restartfile1, maxn=10)
        result2 = sce(rosen, self.x0, self.lb, self.ub, polish=False, seed=1,
                      restart=True)
        assert_equal(result.x, result2.x)
        assert_equal(result.nfev, result2.nfev)

        toremove = [restartfile1, restartfile2]
        for ff in toremove:
            if os.path.exists(ff):
                os.remove(ff)

        # printit
        result = sce(rosen, self.x0, self.lb, self.ub, printit=0)
        assert_equal(result.fun, rosen(result.x))
        result = sce(rosen, self.x0, self.lb, self.ub, printit=1)
        assert_equal(result.fun, rosen(result.x))
        result = sce(rosen, self.x0, self.lb, self.ub, printit=2)
        assert_equal(result.fun, rosen(result.x))
        result = sce(rosen, self.x0, self.lb, self.ub, printit=1, maxn=10)
        assert_equal(result.fun, rosen(result.x))
        result = sce(rosen, self.x0, self.lb, self.ub, printit=2, maxn=10)
        assert_equal(result.fun, rosen(result.x))
        result = sce(rosen, [0.5] * 5, [0.] * 5, [2.] * 5, printit=1,
                     pcento=10)
        print(result.message)
        assert_equal(result.fun, rosen(result.x))

        # sampling
        result = sce(rosen, self.x0, self.lb, self.ub,
                     sampling=['half-open'] * len(self.lb))
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        smpls = ['half-open', 'right-half-open', 'left-half-open', 'open',
                 'log']
        for sm in smpls:
            result = sce(rosen, self.x0, self.lb, self.ub, sampling=sm)
            assert_almost_equal(result.x, self.rosenx, decimal=3)
        lb = [-1, -1.]
        result = sce(rosen, self.x0, lb, self.ub, sampling='log')
        assert_almost_equal(result.x, self.rosenx, decimal=3)
        lb = [0., 0.]
        result = sce(rosen, self.x0, lb, self.ub, sampling='log')
        assert_almost_equal(result.x, self.rosenx, decimal=3)

        # maxit
        result = sce(self.negative_rosen, self.x0, self.lb, self.ub, maxn=100,
                     polish=True, maxit=True)
        assert_equal(result.fun, self.negative_rosen(result.x))

    def test_errors(self):
        # test that the bounds checking works
        func = rosen
        x0 = [1.]
        bounds = [(-3)]
        assert_raises(TypeError, sce, func, x0, bounds)
        bounds = [(-1, 1), (-1, 1)]
        assert_raises(ValueError, sce, func, x0, bounds, sampling='unknown')
        # test correct bool string
        assert_raises(ValueError, _strtobool, 'Ja')
        # test no initial population found
        func = deb03
        x0 = [-0.5, -0.5]
        bounds = [(-1, 0), (-1, 0.)]  # should be (0, 1) to work
        assert_raises(ValueError, sce, func, x0, bounds, printit=1)

    def test_sce_args(self):
        # args
        a = 0.5
        result = sce(rosenbrock_args, self.x0, self.lb, self.ub,
                     args=(a,))
        assert_equal(result.fun, rosenbrock_args(result.x, a))
        assert_almost_equal(result.x, (a, a**2), decimal=3)
        # args and kwargs
        a = 0.5
        b = 200
        result = sce(rosenbrock_args, self.x0, self.lb, self.ub,
                     args=(a,), kwargs={'b': b})
        assert_equal(result.fun, rosenbrock_args(result.x, a, b=b))
        assert_almost_equal(result.x, (a, a**2), decimal=3)
        # kwargs
        a = 0.5
        b = 200
        result = sce(rosenbrock_kwargs, self.x0, self.lb, self.ub,
                     kwargs={'a': a, 'b': b})
        assert_equal(result.fun, rosenbrock_kwargs(result.x, a=a, b=b))
        assert_almost_equal(result.x, (a, a**2), decimal=3)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function

__all__ = ['head']

def head(ifile, n=10, noblank=False, comment='', skip=0, keepnewline=False):
    """
        Return list with first n lines of file.
        Equivalent to Unix head utility.

        If n is omitted, it defaults to 10.

        Blank (only whitespace) and comment lines can be excluded.


        Definition
        ----------
        def head(ifile, n=10, noblank=False, comment='', skip=0, keepnewline=True):


        Input
        -----
        ifile        file handle or filename


        Optional input parameters
        -------------------------
        n            Get the first n lines from file (default: 10)
        comment      Line gets excluded if first character of line is in comment sequence.
                     Sequence can be string, list or tuple.
        skip         Number of lines to skip at the beginning of
                     the file (default 0).


        Options
        -------
        noblank      if True, excludes all lines that consists only of
                     whitespace characters.
        keepnewline  if True, newline character at end of string will not be removed.


        Output
        ------
        list with strings of first n lines in file


        Restrictions
        ------------
        Only ascii files.


        Examples
        --------
        # Create some data
        >>> filename = 'test.dat'
        >>> ifile = open(filename,'w')
        >>> ifile.writelines('First line\\n')
        >>> ifile.writelines(' \\n')
        >>> ifile.writelines('# First comment\\n')
        >>> ifile.writelines('! Second comment\\n')
        >>> ifile.writelines('Last line\\n')
        >>> ifile.close()

        # Read first 10 lines
        >>> print(head(filename))
        ['First line', ' ', '# First comment', '! Second comment', 'Last line']

        # Read first 10 lines, keep newlines in strings
        >>> print(head(filename, keepnewline=True))
        ['First line\\n', ' \\n', '# First comment\\n', '! Second comment\\n', 'Last line\\n']

        # Read first 10 lines excluding blank lines
        >>> print(head(filename, noblank=True))
        ['First line', '# First comment', '! Second comment', 'Last line']

        # Read first 10 lines excluding lines starting with #
        >>> print(head(filename, comment='#'))
        ['First line', ' ', '! Second comment', 'Last line']

        # Read first 10 lines excluding lines starting with # and !
        >>> print(head(filename, comment='#!'))
        ['First line', ' ', 'Last line']

        # Read first 10 lines excluding lines starting with # and S
        >>> print(head(filename, comment='#S'))
        ['First line', ' ', '! Second comment', 'Last line']

        # Read first 10 lines excluding lines starting with # and L
        >>> print(head(filename, comment=('#','L')))
        ['First line', ' ', '! Second comment']

        # Read first 10 lines excluding lines starting with # and ! but comment given as list not string
        >>> print(head(filename, comment=['#','!']))
        ['First line', ' ', 'Last line']

        # Read first 10 lines excluding lines starting with # and ! and blank lines
        >>> print(head(filename, comment='#!', noblank=True))
        ['First line', 'Last line']

        # Read first 10 lines excluding the first 2 lines
        >>> print(head(filename, skip=2))
        ['# First comment', '! Second comment', 'Last line']

        # Read first 3 lines
        >>> print(head(filename, n=3))
        ['First line', ' ', '# First comment']

        # Read first 1 line excluding lines starting with # and ! and blank lines while skipping 1 line
        >>> print(head(filename, 1, skip=1, comment='#!', noblank=True))
        ['Last line']

        # Read first 10 lines from an open file handle rather than a file name
        >>> f = open(filename, 'r')
        >>> print(head(f, 1, skip=1, comment='#!', noblank=True))
        ['Last line']
        >>> f.close()

        # Clean up
        >>> import os
        >>> os.remove(filename)


        Notes
        -----
        if file handle is given, the file pointer will point to the next line after the n-th line read.


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2016 Matthias Cuntz, David Schaefer - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Nov 2014
        Modified, DS, Nov 2014
                  MC, Nov 2016 - adapted file handling to Python 2 and 3
    """
    # Open file
    import sys
    if sys.version_info > (3,0):
        import io
        f = ifile if isinstance(ifile,io.TextIOWrapper) else open(ifile,"r")
    else:
        f = ifile if isinstance(ifile,file) else open(ifile,"r")

    # consume lines
    for _ in range(skip):
        f.readline()

    # Read lines
    count = 0
    header = []
    while count < n:
        line = f.readline()
        if not line:
            break
        if noblank and not line.strip():
            continue
        if line[0] in comment:
            continue
        if not keepnewline:
            line = line.strip("\n")
        header.append(line)
        count += 1

    if f is not ifile:
        f.close()

    return header


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from pyjams.jams.date2dec import date2dec
from pyjams.const import cheat_quartz, cheat_water, cheat_air, density_quartz
from scipy.interpolate import splrep, splint
import scipy.optimize as opt


def energyclosure(fluxfile, metfile, outdir, Rn, G, swdr, Ts=None, theta=None,
                  depths=None, por=None, method='year', force0=True,
                  delimiter=[',',','], skiprows=[1,1], format=['ascii','ascii'],
                  undef=-9999, plot=False):
    '''
    Calculation of energy balance closure and correction of Eddy Covariance data
    with different methods. Possible calculation of soil heat flux from soil
    temperature and moisture if not given.


    Definition
    ----------
    energyclosure(fluxfile, metfile, outdir, Rn, G, swdr, Ts=None, theta=None,
                  depths=None, por=None, method='year', force0=True,
                  delimiter=[',',','], skiprows=[1,1], format=['ascii','ascii'],
                  undef=-9999, plot=False):


    Input
    -----
    fluxfile    str, path and file name of fluxflag or fluxfill or fluxpart
                output file containing fluxes and flags
    metfile     str, path and file name of the meteorology file (must be in
                sync with fluxfile)
    outdir      str, path of the output folder
    Rn          int, column number of net radiation [W m-2] in metfile, column
                number starts with 0 which is first data column.
    G           int, column number of soil heat flux [W m-2] in metfile, column
                number starts with 0 which is first data column. If soil heat
                flux not available, set to False
    swdr        int, column number of short wave downward radiation [W m-2] in
                metfile, column number starts with 0 which is first data column.
                swdr is used for swdr>0=isday


    Optional Input
    --------------
    Ts          list(M) of int, if G is not given, column numbers of soil
                temperatures in certain depths in metfile, column number starts
                with 0 which is first data column.
    theta       list(M) of int, if G is not given, column numbers of soil
                moistures in certain depths in metfile, column number starts
                with 0 which is first data column.
    depths      list(M) or array(M) of int, if G is not given, depths of Ts and
                theta measurements [m], positively increasing with depths
                e.g. [0.05, 0.30, 0.60]
    por         float, if G is not given, porosity of the soil [-], must be
                bigger or equal than np.amax(theta)
    method      str, method of how energy balance closure is calculated and
                applied.
                    if 'year', fit of whole year daytime flag=0 data
                    Rn-G vs. H+Le and application of the fraction to all H, Le
                    and E values
                    implement new methods here
                    implement new methods here
                    implement new methods here
                    (default: 'year')
    force0      bool, if method='year', True forces fit through origin, if False
                fit is allowed to have intercept (default: True)
    delimiter   list of str, delimiters of fluxfile and metfile
                (default: [',',','])
    skiprows    list of int, lines to skip at the beginning of fluxfile and
                metfile, e.g. header lines (default: [1,1])
    format      list of str, time formats of fluxfile and metfile, 'ascii' and
                'eng' possible (default: ['ascii','ascii'])
    undef       int/float, missing value of fluxfile and metfile
                (default: -9999, np.nan is not possible)
    plot        bool, if True performs plotting (default: False)


    Output
    ------
    fluxclosed.csv file containing fluxes and flags where depending on the
                        method chosen fluxes are corrected for energy balance
                        closure
    fluxclosed.pdf if method='year', plot of whole year daytime flag=0 data
                   Rn-G vs. H+Le

    fluxclosed_stor.pdf if method='year', plot of whole year daytime flag=0 data
                   Rn-G vs. H+Le including storages


    Restrictions
    ------------
    Works only with half hourly time steps


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Sep 2014
    '''
    ###########################################################################
    # reading input files
    d = np.loadtxt(fluxfile, dtype='|S100', delimiter=delimiter[0])
    m = np.loadtxt(metfile,  dtype='|S100', delimiter=delimiter[1])

    assert (d.shape[1]==22) | (d.shape[1]==32), 'energyclosure: fluxfile must be from fluxpart and have 22 or 32 cols'

    if format[0]=='ascii':
        datev   = date2dec(ascii=d[skiprows[0]:,0])
    elif format[0]=='eng':
        datev   = date2dec(eng=d[skiprows[0]:,0])
    else:
        raise ValueError('energyclosure: unknown format')
    if format[1]=='ascii':
        datem   = date2dec(ascii=m[skiprows[1]:,0])
    elif format[1]=='eng':
        datem   = date2dec(eng=m[skiprows[1]:,0])
    else:
        raise ValueError('energyclosure: unknown format')

    val = np.where(d[skiprows[0]:,1:]=='', str(undef), d[skiprows[0]:,1:]).astype(np.float)
    met = np.where(m[skiprows[1]:,1:]=='', str(undef), m[skiprows[1]:,1:]).astype(np.float)

    ###########################################################################
    # assign variables
    if (d.shape[1]==22):
        H         = val[:,0]
        Hflag     = val[:,1]
        Le        = val[:,3]
        Leflag    = val[:,4]
        E         = val[:,6]
        Eflag     = val[:,7]
    else:
        H         = val[:,0]
        Hflag     = val[:,2]
        Le        = val[:,4]
        Leflag    = val[:,6]
        E         = val[:,8]
        Eflag     = val[:,10]
        H_stor      = val[:,1]
        Hflag_stor  = val[:,2]
        Le_stor     = val[:,5]
        Leflag_stor = val[:,6]
        E_stor      = val[:,9]
        Eflag_stor  = val[:,10]
    Rn        = met[:,Rn]
    Ts        = met[:,Ts]
    swdr      = met[:,swdr]
    isday     = swdr>0.

    ###########################################################################
    # check if soil heat flux is given or needs to be calculated
    if not G:
        if Ts==None or theta==None or depths==None or por==None:
            raise ValueError('energyclosure: if G is not given, Ts, theta, depths and rhos are needed to calculate G')
        else:
            G=soilheatflux(Ts, theta, depths, rhos)
    else:
        G = met[:,G]

    ###########################################################################
    # apply energy balance closure methods
    # yearly approach
    if (method.lower() == 'year'):
        closure = energyclosure_year(H, Hflag, Le, Leflag, Rn, G, isday, outdir,
                                     force0=force0, undef=undef, plot='energyclosure.pdf')
        if force0:
            H_closed, Le_closed, E_closed = H/closure, Le/closure, E/closure
        else:
            H_closed, Le_closed, E_closed = H/closure[0], Le/closure[0], E/closure[0]

        if (d.shape[1]==32):
            closure_stor = energyclosure_year(H_stor, Hflag_stor, Le_stor, Leflag_stor,#
                                              Rn, G, isday, outdir, force0=force0,
                                              undef=undef, plot='energyclosure_stor.pdf')
            if force0:
                H_stor_closed, Le_stor_closed, E_stor_closed = H_stor/closure, Le_stor/closure, E_stor/closure
            else:
                H_stor_closed, Le_stor_closed, E_stor_closed = H_stor/closure[0], Le_stor/closure[0], E_stor/closure[0]

    # Include new methods here
    # Include new methods here
    # Include new methods here
    else:
        raise ValueError('energyclosure: method not implemented yet.')

    H_closed[H==undef]   = undef
    Le_closed[Le==undef] = undef
    E_closed[E==undef]   = undef
    if (d.shape[1]==32):
        H_stor_closed[H_stor==undef]   = undef
        Le_stor_closed[Le_stor==undef] = undef
        E_stor_closed[E_stor==undef]   = undef

    ###########################################################################
    # prepare and write output
    if (d.shape[1]==22):
        flux = np.vstack((H_closed,Le_closed,E_closed)).transpose()
    else:
        flux = np.vstack((H_closed,H_stor_closed,Le_closed,Le_stor_closed,E_closed,E_stor_closed)).transpose()

    flux_str = np.array([['%11.5f'%x for x in y] for y in flux])
    flux_str = np.where(flux_str=='%11.5f'%undef, ' '*(11-len(str(undef)))+str(undef), flux_str)

    if (d.shape[1]==22):
        d[skiprows[0]:,[1,4,7]] = flux_str
    else:
        d[skiprows[0]:,[1,2,5,6,9,10]] = flux_str

    np.savetxt('%s/fluxclosed.csv'%outdir, d, '%s', delimiter=',')

def soilheatflux(Ts, theta, depths, por, undef=-9999):
    '''
    Calculates soil heat flux from soil temperatures and soil moistures at
    different depths. It's best do include surface temperatures at 0 m.


    Definition
    ----------
    soilheatflux(Ts, theta, depths, por, undef=-9999):


    Input
    -----
    Ts          array(N,M), soil temperatures [K or degC]
    theta       array(N,M), soil moisture [-]
    depths      list(M) or array(M), depths of the Ts and theta measurements [m]
                positively increasing with depths e.g. [0.05, 0.30, 0.60]
    por         float, porosity of the soil [-] must be bigger or equal
                than np.amax(theta)


    Optional Input
    --------------
    undef       int/float, missing value of Ts and theta
                (default: -9999, np.nan is not possible)


    Output
    ------
    G           array(N), soil heat flux [W m-2]


    Restrictions
    ------------
    Works only with half hourly time steps


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Sep 2014
    '''
    if (np.size(depths)!=Ts.shape[1]) or (Ts.shape!=theta.shape):
        raise ValueError('soilheatflux: dimensions mismatch')

    Ts    = np.ma.array(Ts, mask=Ts==undef)
    theta = np.ma.array(theta, mask=theta==undef)

    # calc soil density and soil heat capacity
    rhos  = (1.-por)*density_quartz
    csoil = cheat_quartz*(1.-por) + cheat_water*theta + cheat_air*(1-theta)*por

    # calculate heat flux
    T_diff = (Ts[:-1,:] - Ts[1:,:])/1800. * rhos * csoil[:-1,:]
    G      = np.ma.masked_all_like(Ts[:,0])
    for i, item in enumerate(T_diff):
        if not item.mask.any():
            tck  = splrep(depths, item, k=1)
            G[i+1] = splint(depths[0],depths[-1],tck)
    G[0]=G[1]

    return np.ma.filled(G, fill_value=undef)

def energyclosure_year(H, Hflag, Le, Leflag, Rn, G, isday, outdir, force0=True,
                       undef=-9999, plot=False):
    '''
    Calculates energy balance closure with the 'year' approach. A straight line
    is fitted to Rn-g vs. H+Le and the fraction is returned.


    Definition
    ----------
    energyclosure_year(H, Le, Rn, G, isday, outdir, force0=force0, undef=undef,
                       plot=plot):


    Input
    -----
    H           array(N), sensible heat flux [W m-2]
    Hflag       array(N), quality flag of sensible heat flux, 0 where good quality
    Le          array(N), latent heat flux [W m-2]
    Leflag      array(N), quality flag of latent heat flux, 0 where good quality
    Rn          array(N), net radiation [W m-2]
    G           array(N), soil heat flux [W m-2]
    isday       array(N), bool, True if day, False if night, fitting is done
                only to day data
    outdir      str, path of the output folder


    Optional Input
    --------------
    force0      bool, if True fitting line is forced to origin and output is
                only the fraction, if False fitted line can have a intercept and
                output is then (fraction, intercept), (default: True)
    undef       int/float, missing value of input (default: -9999, np.nan is
                not possible)
    plot        bool, if False, no fitting plot is made and saved to outdir
                (default: False), otherwise give file name like 'energyclosure.pdf'


    Output
    ------
    fraction    float, energy balance closure gap, e.g 0.8 means 80% of
                available energy is observed, 20% are missing. list(2) of floats
                with energy balance closure gap and intercept if force0=False


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Sep 2014
    '''
    ###########################################################################
    # variables
    H  = np.ma.array(H,  mask=(H==undef)  | (~isday) |(Hflag>0))
    Le = np.ma.array(Le, mask=(Le==undef) | (~isday) |(Leflag>0))
    Rn = np.ma.array(Rn, mask=(Rn==undef) | (~isday))
    G  = np.ma.array(G,  mask=(G==undef)  | (~isday))

    x = Rn-G
    y = H+Le

    ###########################################################################
    # fit
    if force0:
        p_guess = [1]
        func    = lin2
    else:
        p_guess = [0,1]
        func    = lin

    if plot:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf
        x_mod = np.arange(np.ma.min(x),np.ma.max(x),(np.ma.max(x)-np.ma.min(x))/50.)
        fig1=plt.figure('pre-fit')
        fig1.add_subplot(111, aspect='equal')
        plt.axhline(y=0, color='k')
        plt.axvline(x=0, color='k')
        plt.plot(x, y, 'bo', label='n: %i'%np.ma.count(x+y))
        plt.plot(x_mod, func(x_mod,p_guess), 'r-')
        plt.xlabel('Rn-G')
        plt.ylabel('H+Le')
        plt.legend(loc='best')
        plt.show()

    p_opt = opt.fmin(absdif, p_guess, args=(x, y, func), disp=0)

    if plot:
        fig1=plt.figure('post-fit')
        fig1.add_subplot(111, aspect='equal')
        plt.axhline(y=0, color='k')
        plt.axvline(x=0, color='k')
        plt.plot(x, y, 'bo', label='n: %i'%np.ma.count(x+y))
        if force0:
            plt.plot(x_mod, func(x_mod,p_opt), 'r-', label='slope: %0.2f'%p_opt)
        else:
            plt.plot(x_mod, func(x_mod,p_opt), 'r-', label='slope: %0.2f'%p_opt[0])
        plt.xlabel('Rn-G')
        plt.ylabel('H+Le')
        plt.legend(loc='best')
        plt.show()

        pp1 = pdf.PdfPages('%s/'%(outdir)+plot)
        fig1.savefig(pp1, format='pdf')
        pp1.close()

    return p_opt

def absdif(p,x,y,func):
    '''
    energyclosure: objective function
    '''
    return np.ma.sum(np.ma.abs(y-func(x, p)))

def lin(x, p):
    '''
    energyclosure: linear function
    '''
    return p[0] * x + p[1]

def lin2(x, p):
    '''
    energyclosure: linear function with no intercept
    '''
    return p[0] * x

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
This is the unittest for Screening module.

python -m unittest -v tests/test_screening.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_screening.py

"""
import unittest


class TestScreening(unittest.TestCase):
    """
    Tests for screening.py
    Missing coverage:
        ca. 230: using MPI
        ca. 265: no seed given
        ca. 310: only one trajectory
    """
    def setUp(self):
        import numpy as np
        # seed for reproducible results
        seed = 1234
        np.random.seed(seed=seed)
        self.nt      = 10
        self.ntotal  = 50
        self.nsteps  = 6
        self.verbose = 1

    # G function
    def test_ee_g(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        out = ee(obj, lb, ub, self.nt, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=1)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.045, 0.24, 1.624, 0.853, 0.031, 0.084])

    # G function
    def test_ee_g_verbose(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        out = ee(obj, lb, ub, self.nt, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=1, verbose=1)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.045, 0.24, 1.624, 0.853, 0.031, 0.084])

    # G function, mask
    def test_ee_g_mask(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb   = np.zeros(npars)
        ub   = np.ones(npars)
        x0   = np.ones(npars)*0.5
        mask = np.ones(npars, dtype=bool)
        mask[1] = False

        out = ee(obj, lb, ub, self.nt, x0=x0, mask=mask,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=1)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.038, 0.0, 1.582, 1.136, 0.04, 0.102])

    # G function, mask, error
    def test_ee_g_mask_error(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb   = np.zeros(npars)
        ub   = np.ones(npars)
        x0   = np.ones(npars)*0.5
        mask = np.ones(npars, dtype=bool)
        mask[1] = False

        self.assertRaises(TypeError, ee, obj, lb, ub, self.nt, x0=None,
                          mask=mask, ntotal=self.ntotal, nsteps=self.nsteps,
                          processes=1)

    # G function, nt=1
    def test_ee_g_nt1(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        out = ee(obj, lb, ub, 1, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=1)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.0, 0.134, 1.813, 0.0, 0.017, 0.0])

    # G function, pool
    def test_ee_g_pool(self):
        import numpy as np
        import schwimmbad
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        nprocs = 4
        ipool = schwimmbad.choose_pool(mpi=False, processes=nprocs)
        out = ee(obj, lb, ub, self.nt, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=nprocs, pool=ipool)
        ipool.close()

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.045, 0.24, 1.624, 0.853, 0.031, 0.084])

    # G function, multiprocesses
    def test_ee_g_1(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import G
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = G
        npars  = 6
        params = [78., 12., 0.5, 2., 97., 33.]  # G

        # Partialise function with fixed parameters
        arg   = [params]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        out = ee(obj, lb, ub, npars, processes=4)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.047, 0.233, 1.539, 0.747, 0.025, 0.077])

    # Gstar function with different interactions
    def test_screening_gstar(self):
        import numpy as np
        from pyjams import screening
        from pyjams.functions import Gstar
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func   = Gstar
        npars  = 10
        params = [[np.ones(npars),     np.random.random(npars),
                   [0., 0.,  9.,  9.,  9.,  9.,  9., 9., 9., 9.]],  # G*
                  [np.ones(npars),     np.random.random(npars),
                   [0., 0.1, 0.2, 0.3, 0.4, 0.8, 1., 2., 3., 4.]],
                  [np.ones(npars)*0.5, np.random.random(npars),
                   [0., 0.,  9.,  9.,  9.,  9.,  9., 9., 9., 9.]],
                  [np.ones(npars)*0.5, np.random.random(npars),
                   [0., 0.1, 0.2, 0.3, 0.4, 0.8, 1., 2., 3., 4.]],
                  [np.ones(npars)*2.0, np.random.random(npars),
                   [0., 0.,  9.,  9.,  9.,  9.,  9., 9., 9., 9.]],
                  [np.ones(npars)*2.0, np.random.random(npars),
                   [0., 0.1, 0.2, 0.3, 0.4, 0.8, 1., 2., 3., 4.]]]
        iiout  = [[1.087, 1.807, 0.201, 0.13,  0.08,
                   0.077, 0.055, 0.198, 0.139, 0.136],
                  [0.924, 1.406, 0.603, 0.876, 1.194,
                   0.567, 0.642, 0.276, 0.131, 0.34 ],
                  [1.021, 0.875, 0.085, 0.096, 0.116,
                   0.104, 0.11,  0.07,  0.045, 0.112],
                  [1.096, 0.752, 0.573, 0.762, 0.189,
                   0.584, 0.259, 0.437, 0.468, 0.169],
                  [4.806, 3.299, 0.614, 0.406, 0.89,
                   0.628, 0.372, 0.428, 0.64,  0.565],
                  [1.212, 1.076, 3.738, 8.299, 0.636,
                   0.469, 0.37,  4.856, 0.553, 0.009]]
        lb = np.zeros(npars)
        ub = np.ones(npars)

        for ii in range(len(params)):
            # Partialise function with fixed parameters
            arg   = params[ii]
            kwarg = {}
            obj   = _FunctionWrapper(func, arg, kwarg)

            out = screening(obj, lb, ub, self.nt, x0=None, mask=None,
                            ntotal=self.ntotal, nsteps=self.nsteps,
                            processes=1)  # , plotfile='gstar'+str(ii)+'.png')
            # Check
            self.assertEqual(list(np.around(out[:, 0], 3)), iiout[ii])

    # Bratley / K function
    def test_ee_k(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import bratley

        # Function and parameters
        func   = bratley
        npars  = 10
        params = []  # k

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        out = ee(func, lb, ub, self.nt, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=1)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.586, 0.219, 0.082, 0.055, 0.02,
                          0.068, 0.009, 0.007, 0., 0.])

    # Morris function
    def test_ee_fmorris(self):
        import numpy as np
        from pyjams import ee
        from pyjams.functions import fmorris
        from pyjams.sce import _FunctionWrapper

        # Function and parameters
        func = fmorris
        npars = 20
        beta0                 = 0.
        beta1                 = np.random.standard_normal(npars)
        beta1[:10]            = 20.
        beta2                 = np.random.standard_normal((npars, npars))
        beta2[:6, :6]         = -15.
        beta3                 = np.zeros((npars, npars, npars))
        beta3[:5, :5, :5]     = -10.
        beta4                 = np.zeros((npars, npars, npars, npars))
        beta4[:4, :4, :4, :4] = 5.

        # Partialise Morris function with fixed parameters beta0-4
        arg   = [beta0, beta1, beta2, beta3, beta4]
        kwarg = {}
        obj   = _FunctionWrapper(func, arg, kwarg)

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)

        # Check
        out = ee(obj, lb, ub, self.nt, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 processes=4)

        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [66.261, 48.363, 20.133, 62.42, 37.87,
                          26.031, 30.211, 39.714, 42.633, 43.776,
                          4.996, 3.701, 4.734, 8.031, 5.734,
                          3.564, 5.068, 7.635, 3.129, 5.224])

    # Bratley / K function with distribution
    def test_ee_k_dist(self):
        import numpy as np
        import scipy.stats as stats
        from pyjams import ee
        from pyjams.functions import bratley

        # Function and parameters
        func   = bratley
        npars  = 10
        params = []  # k

        # Screening
        lb = np.zeros(npars)
        ub = np.ones(npars)
        dist      = [ stats.uniform for i in range(npars) ]
        distparam = [ (lb[i], ub[i]-lb[i]) for i in range(npars) ]
        lb = np.zeros(npars)
        ub = np.ones(npars)

        out = ee(func, lb, ub, self.nt, x0=None, mask=None,
                 ntotal=self.ntotal, nsteps=self.nsteps,
                 dist=dist, distparam=distparam,
                 processes=1)

        # Check
        self.assertEqual(list(np.around(out[:, 0], 3)),
                         [0.586, 0.219, 0.082, 0.055, 0.02,
                          0.068, 0.009, 0.007, 0., 0.])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
This is the unittest for the Fit Functions module.

python -m unittest -v tests/test_fit_functions.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_fit_functions.py

"""
import unittest


class TestFitFunctions(unittest.TestCase):
    """
    Tests for functions/logistic_function.py
    Missing coverage:
        ca. 200ff: exception for import at different levels

    """

    def test_cost_abs_square(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import logistic_p, cost_logistic, cost2_logistic
        from pyjams.functions import cost_abs, cost_square

        p = [1., 1., 0.]
        x = np.arange(2)
        y = np.zeros(2)
        assert cost_logistic(p, x, y) == cost_abs(p, logistic_p, x, y)
        assert cost2_logistic(p, x, y) == cost_square(p, logistic_p, x, y)
        assert (cost_logistic(p, pd.Series(x), pd.Series(y)) ==
                cost_abs(p, logistic_p, x, y))

    def test_arrhenius(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import arrhenius, arrhenius_p, cost_arrhenius
        from pyjams.functions import cost2_arrhenius

        T25 = 298.15
        R   = 8.31446261815324
        p = [1.]
        x = np.arange(2) + 25.
        y = np.zeros(2)
        out = np.exp(np.arange(2) / (T25 * R * (np.arange(2) + T25)))
        self.assertEqual(list(arrhenius(x, *p)), list(out))
        self.assertEqual(list(arrhenius_p(x, p)), list(out))
        self.assertEqual(list(arrhenius(pd.Series(x), *p)), list(out))
        self.assertEqual(list(arrhenius_p(pd.Series(x), p)), list(out))
        assert cost_arrhenius(p, x, y) == np.sum(out)
        assert cost2_arrhenius(p, x, y) == np.sum(out**2)

    def test_f1x(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import f1x, f1x_p, cost_f1x, cost2_f1x

        p = [0., 2.]
        x = np.arange(2) + 1.
        y = np.zeros(2)
        out = 2. / x
        self.assertEqual(list(f1x(x, *p)), list(out))
        self.assertEqual(list(f1x_p(x, p)), list(out))
        self.assertEqual(list(f1x(pd.Series(x), *p)), list(out))
        self.assertEqual(list(f1x_p(pd.Series(x), p)), list(out))
        assert cost_f1x(p, x, y) == np.sum(out)
        assert cost2_f1x(p, x, y) == np.sum(out**2)

    def test_fexp(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import fexp, fexp_p, cost_fexp, cost2_fexp

        p = [0., 1., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = np.exp(x)
        self.assertEqual(list(fexp(x, *p)), list(out))
        self.assertEqual(list(fexp_p(x, p)), list(out))
        self.assertEqual(list(fexp(pd.Series(x), *p)), list(out))
        self.assertEqual(list(fexp_p(pd.Series(x), p)), list(out))
        assert cost_fexp(p, x, y) == np.sum(out)
        assert cost2_fexp(p, x, y) == np.sum(out**2)

    def test_fgauss(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import gauss, gauss_p, cost_gauss, cost2_gauss

        p = [1., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / np.sqrt(2. * np.pi) * np.exp(-np.abs(x - 1.) / 2.)
        self.assertEqual(list(gauss(x, *p)), list(out))
        self.assertEqual(list(gauss_p(x, p)), list(out))
        self.assertEqual(list(gauss(pd.Series(x), *p)), list(out))
        self.assertEqual(list(gauss_p(pd.Series(x), p)), list(out))
        assert cost_gauss(p, x, y) == np.sum(out)
        assert cost2_gauss(p, x, y) == np.sum(out**2)

    def test_lasslop(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import lasslop, lasslop_p, cost_lasslop
        from pyjams.functions import cost2_lasslop

        p = [1., 1., 1., 1.]
        Rg = np.arange(2) * 1000.
        et = np.arange(2)
        VPD = np.arange(2) * 10. + 1000.
        y = np.zeros(2)
        beta = np.array([1., np.exp(-10.)])
        out = et - beta * Rg / (Rg + beta)
        self.assertEqual(list(lasslop(Rg, et, VPD, *p)), list(out))
        self.assertEqual(list(lasslop_p(Rg, et, VPD, p)), list(out))
        df = pd.DataFrame({'Rg': Rg, 'et': et, 'VPD': VPD})
        self.assertEqual(list(lasslop(df['Rg'], df['et'], df['VPD'], *p)),
                         list(out))
        self.assertEqual(list(lasslop_p(df['Rg'], df['et'], df['VPD'], p)),
                         list(out))
        assert cost_lasslop(p, Rg, et, VPD, y) == np.sum(out)
        assert cost2_lasslop(p, Rg, et, VPD, y) == np.sum(out**2)

    def test_line(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import line, line_p, cost_line, cost2_line

        p = [0., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = np.arange(2)
        self.assertEqual(list(line(x, *p)), list(out))
        self.assertEqual(list(line_p(x, p)), list(out))
        self.assertEqual(list(line(pd.Series(x), *p)), list(out))
        self.assertEqual(list(line_p(pd.Series(x), p)), list(out))
        assert cost_line(p, x, y) == np.sum(out)
        assert cost2_line(p, x, y) == np.sum(out**2)

    def test_line0(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import line0, line0_p, cost_line0, cost2_line0

        p = [1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = np.arange(2)
        self.assertEqual(list(line0(x, *p)), list(out))
        self.assertEqual(list(line0_p(x, p)), list(out))
        self.assertEqual(list(line0(pd.Series(x), *p)), list(out))
        self.assertEqual(list(line0_p(pd.Series(x), p)), list(out))
        assert cost_line0(p, x, y) == np.sum(out)
        assert cost2_line0(p, x, y) == np.sum(out**2)

    def test_lloyd_fix(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import lloyd_fix, lloyd_fix_p, cost_lloyd_fix
        from pyjams.functions import cost2_lloyd_fix

        p = [1., 1.]
        x = np.arange(2) + 273.15
        y = np.zeros(2)
        out = np.exp(1. / 56.02 - 1. / (x - 227.13))
        self.assertEqual(list(lloyd_fix(x, *p)), list(out))
        self.assertEqual(list(lloyd_fix_p(x, p)), list(out))
        self.assertEqual(list(lloyd_fix(pd.Series(x), *p)), list(out))
        self.assertEqual(list(lloyd_fix_p(pd.Series(x), p)), list(out))
        assert cost_lloyd_fix(p, x, y) == np.sum(out)
        assert cost2_lloyd_fix(p, x, y) == np.sum(out**2)

    def test_lloyd_only_rref(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import lloyd_only_rref, lloyd_only_rref_p
        from pyjams.functions import cost_lloyd_only_rref
        from pyjams.functions import cost2_lloyd_only_rref

        p = [2.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 2. * np.arange(2)
        self.assertEqual(list(lloyd_only_rref(x, *p)), list(out))
        self.assertEqual(list(lloyd_only_rref_p(x, p)), list(out))
        self.assertEqual(list(lloyd_only_rref(pd.Series(x), *p)), list(out))
        self.assertEqual(list(lloyd_only_rref_p(pd.Series(x), p)), list(out))
        assert cost_lloyd_only_rref(p, x, y) == np.sum(out)
        assert cost2_lloyd_only_rref(p, x, y) == np.sum(out**2)

    def test_logistic(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import logistic, logistic_p, cost_logistic
        from pyjams.functions import cost2_logistic

        p = [1., 1., 0.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / (1. + np.exp(-x))
        self.assertEqual(list(logistic(x, *p)), list(out))
        self.assertEqual(list(logistic_p(x, p)), list(out))
        self.assertEqual(list(logistic(pd.Series(x), *p)), list(out))
        self.assertEqual(list(logistic_p(pd.Series(x), p)), list(out))
        assert cost_logistic(p, x, y) == np.sum(out)
        assert cost2_logistic(p, x, y) == np.sum(out**2)

    def test_logistic_offset(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import logistic_offset, logistic_offset_p
        from pyjams.functions import cost_logistic_offset
        from pyjams.functions import cost2_logistic_offset

        p = [1., 1., 0., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / (1. + np.exp(-x)) + 1.
        self.assertEqual(list(logistic_offset(x, *p)), list(out))
        self.assertEqual(list(logistic_offset_p(x, p)), list(out))
        self.assertEqual(list(logistic_offset(pd.Series(x), *p)), list(out))
        self.assertEqual(list(logistic_offset_p(pd.Series(x), p)), list(out))
        assert cost_logistic_offset(p, x, y) == np.sum(out)
        assert cost2_logistic_offset(p, x, y) == np.sum(out**2)

    def test_logistic2_offset(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import logistic2_offset, logistic2_offset_p
        from pyjams.functions import cost_logistic2_offset
        from pyjams.functions import cost2_logistic2_offset

        p = [1., 1., 0., 1., 2., 0., 1.]
        x = np.arange(2)
        y = np.zeros(2)
        out = 1. / (1. + np.exp(-x)) - 1. / (1. + np.exp(-2. * x)) + 1.
        self.assertEqual(list(np.around(logistic2_offset(x, *p), 14)),
                         list(np.around(out, 14)))
        self.assertEqual(list(np.around(logistic2_offset_p(x, p), 14)),
                         list(np.around(out, 14)))
        self.assertEqual(list(np.around(logistic2_offset(pd.Series(x), *p),
                                        14)), list(np.around(out, 14)))
        self.assertEqual(list(np.around(logistic2_offset_p(pd.Series(x), p),
                                        14)), list(np.around(out, 14)))
        assert (np.around(cost_logistic2_offset(p, x, y), 14) ==
                np.around(np.sum(out), 14))
        assert (np.around(cost2_logistic2_offset(p, x, y), 14) ==
                np.around(np.sum(out**2), 14))

    def test_poly(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import poly, poly_p, cost_poly, cost2_poly

        p = [0., 1., 2., 3]
        x = np.arange(2)
        y = np.zeros(2)
        out = 0. + 1. * x + 2. * x**2 + 3. * x**3
        self.assertEqual(list(poly(x, *p)), list(out))
        self.assertEqual(list(poly_p(x, p)), list(out))
        self.assertEqual(list(poly(pd.Series(x), *p)), list(out))
        self.assertEqual(list(poly_p(pd.Series(x), p)), list(out))
        assert cost_poly(p, x, y) == np.sum(out)
        assert cost2_poly(p, x, y) == np.sum(out**2)

    def test_sabx(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import sabx, sabx_p, cost_sabx, cost2_sabx

        p = [0., 2.]
        x = np.arange(2) + 1.
        y = np.zeros(2)
        out = np.sqrt(2. / x)
        self.assertEqual(list(sabx(x, *p)), list(out))
        self.assertEqual(list(sabx_p(x, p)), list(out))
        self.assertEqual(list(sabx(pd.Series(x), *p)), list(out))
        self.assertEqual(list(sabx_p(pd.Series(x), p)), list(out))
        assert cost_sabx(p, x, y) == np.sum(out)
        assert cost2_sabx(p, x, y) == np.sum(out**2)

    def test_see(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import see, see_p, cost_see, cost2_see

        p = [1., 1., 2.]
        x = np.arange(2) + 0.5
        y = np.zeros(2)
        out = np.array([0., 0.25])
        self.assertEqual(list(see(x, *p)), list(out))
        self.assertEqual(list(see_p(x, p)), list(out))
        self.assertEqual(list(see(pd.Series(x), *p)), list(out))
        self.assertEqual(list(see_p(pd.Series(x), p)), list(out))
        assert cost_see(p, x, y) == np.sum(out)
        assert cost2_see(p, x, y) == np.sum(out**2)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
'''
    leafmodel: contains routines to model photosynthesis and stomatal conductance
    of canopies or leaves.
    
    
    major functions contained:
    --------------------------
    opt_leafmodel:  optimization against observed transpiration and
                    gross primary production
    cost_leafmodel: cost function for optimization
    leafmodel:      model for photosynthesis and stomatal conductance
    twoleafmodel:   uses leafmodel in a sunlit and shaded scheme
    farquhar:       photosysnthesis model
    leuning:        stomatal conductance modeling
    ball_berry:     stomatal conductance modeling


    Examples
    --------
    ci, gcmol_c, gcmol_h, gsmol_c, gsmol_h, no_conv, Jc, Je, Rd, GPPmod, Emod = jams.leafmodel.leafmodel(
        ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2, gs_method=gs_method, temp_method=temp_method, n=n, eta=eta)


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP & MC, Jul 2014
'''
from .leafmodel import *

# Information
__author__   = "Arndt Piayda, Matthias Cuntz"
__version__  = '1.0'
__revision__ = "Revision: 1796"
__date__     = 'Date: 05.10.2014'
